from __future__ import annotations
from typing import Any, Optional

from ..enums import DefaultAvatar
from ..asset import Asset
from ..flags import UserFlags, GuildMemberFlags
from ..permissions import Permissions
from pydantic import BaseModel

__all__ = ("User", "Member")


class User(BaseModel):
    id: int
    username: str
    discriminator: str
    avatar: Optional[str]
    avatar_decoration: Optional[str]
    bot: Optional[bool]
    bio: Optional[str]
    system: Optional[bool]
    mfa_enabled: Optional[bool]
    banner: Optional[str]
    ancent_color: Optional[int]
    locale: Optional[str]
    verified: Optional[bool]
    email: Optional[str]
    flags: Optional[UserFlags]
    premium_type: Optional[int]
    public_flags: Optional[int]

    def __str__(self):
        return f"{self.username}#{self.discriminator}"

    @classmethod
    def from_resolved_data(cls, **kwargs):
        return cls(**kwargs)

    @property
    def mention(self) -> str:
        return f"<@{self.id}>"

    @property
    def display_avatar(self) -> Asset:
        if not self.avatar:
            return Asset.from_default_avatar(
                int(self.discriminator) % len(DefaultAvatar)
            )
        return Asset.from_avatar(self.avatar, self.id)


class Member(BaseModel):
    user: Optional[User]
    nick: Optional[str]
    avatar: Optional[str]
    roles: list[Any]
    joined_at: Any
    premium_since: Optional[Any]
    deaf: bool
    mute: bool
    flags: Optional[GuildMemberFlags]
    pending: Optional[bool]
    permissions: Optional[Permissions]
    communication_disabled_until: Optional[Any]

    def __str__(self) -> str:
        return f"{self.user}"



from .client import *
from .models import *
from .enums import *
from .types import *
from .permissions import *
from .plugin import *
from .errors import *
from .commands import *


__version__ = "0.0.17"



from __future__ import annotations
import asyncio
from functools import wraps
from typing import Any, List, Dict, Optional, TYPE_CHECKING, Union
from fastapi import FastAPI
from .models import Guild
from .handler import InteractionHandler
from .http import HttpClient
from .models import User
from .errors import CommandInvokeError
from .commands import Command, Group
from logging import getLogger
from .utils import LOGGING_CONFIG

if TYPE_CHECKING:
    from .ui import View, Component, Modal
    from .types import AsyncFunction
    from .permissions import Permissions
    from .models import Interaction


log = getLogger("dismake")
__all__ = ("Bot",)


class Bot(FastAPI):
    """
    Represents a Discord bot.

    This class is a subclass of `FastAPI`, which means you can create API routes
    and do whatever you can do with `FastAPI`with this class's instance.

    Parameters
    ----------
    token: :class:`str`
        The token for the Discord bot.
    client_public_key: :class:`str`
        The public key for the Discord client.
    client_id: :class:`int`
        The ID for the Discord client.
    route: :class:`str`
        The route to listen for Discord interactions on, by default "/interactions".
    interaction_handler: :class:`InteractionHandler`
        An interaction handler to process incoming Discord interactions, by default :class:`InteractionHandler`.

    Attributes
    ----------
    user: :class:`User`
        The user within this bot.
    """

    def __init__(
        self,
        token: str,
        client_public_key: str,
        client_id: int,
        route: str = "/interactions",
        interaction_handler: Optional[InteractionHandler] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self._client_id = client_id
        self._client_public_key = client_public_key
        self._interaction_handler = interaction_handler or InteractionHandler(self)
        self._http = HttpClient(token=token, client_id=client_id)
        self.add_route(
            path=route,
            route=self._interaction_handler.handle_interactions,
            methods=["POST"],
            include_in_schema=False,
        )
        self.add_event_handler("startup", self._http.fetch_me)
        self._events: Dict[str, List[AsyncFunction]] = {}
        self.add_event_handler("startup", lambda: self.dispatch("ready"))
        self._components: Dict[str, Component] = {}
        self._commands: Dict[str, Union[Group, Command]] = {}
        self._modals: Dict[str, Modal] = {}
        self.error_handler: Optional[AsyncFunction] = None
        self.log = log

    @property
    def user(self) -> User:
        """
        Returns the user object representing the bot.

        This property returns a `User` object that represents the bot in the Discord API.

        Returns
        -------
        user: :class:`User`
            The `User` object representing the bot.
        """
        return self._http._user

    def get_command(self, name: str) -> Optional[Union[Command, Group]]:
        """
        Returns the slash command with the specified name, or None if it doesn't exist.

        Parameters
        ----------
        name: str
            The name of the command to retrieve.

        Returns
        -------
        command: Optional[Union[:class:`Command`, :class:`Group`]]
            The command object with the specified name, or None if no such object exists.
        """
        return self._commands.get(name)

    def run(self, **kwargs):
        """
        Starts the web server to handle HTTP interactions with Discord.

        This method starts a web server using `uvicorn`, which handles HTTP interactions with Discord.
        Any additional keyword arguments are passed directly to `uvicorn.run()`.
        By default, this method uses the `LOGGING_CONFIG` configuration for logging,
        but you can provide your own configuration using the `log_config` keyword argument.
        """
        import uvicorn

        kwargs["log_config"] = kwargs.get("log_config", LOGGING_CONFIG)
        uvicorn.run(**kwargs)

    async def _dispatch_callback(self, coro: AsyncFunction, *args, **kwargs):
        """
        Dispatches an event to a single event listener.

        This method is used internally to dispatch an event to a single event listener.
        It calls the specified coroutine function with the given arguments and keyword arguments.
        If the coroutine raises an exception, the exception is logged using the `log` module's `error()` method.

        Parameters
        ----------
        coro:
            The coroutine function to call.
        """
        try:
            await coro(*args, **kwargs)
        except Exception as e:
            log.error("An error occured in %s" % coro.__name__, exc_info=e)

    def dispatch(self, event_name: str, *args, **kwargs):
        """
        Dispatches an event to all registered event listeners.

        Parameters
        ----------
        event_name: :class:`str`
            The name of the event to dispatch.
        *args: Any
            positional arguments to pass to the event listeners.
        **kwargs: Any
            keyword arguments to pass to the event listeners.
        """
        if not event_name.startswith("on_"):
            event_name = "on_" + event_name
        event = self._events.get(event_name)
        if not event:
            return
        for coro in event:
            asyncio.ensure_future(self._dispatch_callback(coro, *args, **kwargs))

    def event(self, event_name: str | None = None):
        """
        A decorator that registers an event to listen to.

        Parameters
        ----------
        event_name: str
            The event name you want to listen.

        Example usage
        -------------
            >>> import dismake
            >>> app = dismake.Bot(...)
            >>> @app.event()
            ... async def on_ready():
            ...     print(f"Logged in as {app.user}.")
        """

        def decorator(coro: AsyncFunction):
            @wraps(coro)
            def wrapper(*_, **__):
                name = event_name or coro.__name__
                if self._events.get(name) is not None:
                    self._events[name].append(coro)
                else:
                    self._events[name] = [coro]

            return wrapper()

        return decorator

    async def sync_commands(self, guild_ids: Optional[int] = None):
        """
        Synchronizes all application commands to Discord.

        Parameters
        ----------
        guild_ids: Optional[int]
            An optional list of guild IDs to sync commands for. If not specified, commands will be synced globally.

        Returns
        -------
        A list of the updated application commands on success.

        Example usage
        -------------
            >>> from dismake import Dismake
            >>> app = Dismake()
            >>> @app.command()
            ... async def hello(ctx):
            ...     await ctx.respond("Hello, world!")
            >>> @app.event('ready')
            ... async def on_ready():
            ...     await app.sync_commands()
        """
        return await self._http.bulk_override_commands(
            [command for command in self._commands.values()]
        )

    def on_error(self, coro: AsyncFunction):
        """
        A decorator which overrides the `:meth: _default_error_handler`.

        Example usage
        -------------
            >>> import dismake
            >>> app = dismake.Bot(...)
            >>> @app.on_app_command_error
            ... async def custom_error_handler(ctx, error):
            ...     # Your error handler logics
            ...     pass
        """

        @wraps(coro)
        def wrapper(*_, **__):
            self.error_handler = coro
            return coro

        return wrapper()

    async def on_command_error(self, interaction: Interaction, error: Exception) -> Any:
        """
        A default error handler which handles the CommandInvokeError
        """
        if isinstance(error, CommandInvokeError):
            await interaction.send(
                f"Oops! Something went wrong while running the command.", ephemeral=True
            )
        raise error

    async def fetch_guild(self, guild_id: int) -> Guild:
        """
        Fetches a guild from discord by its ID.

        Parameters
        ----------
        guild_id: int
            The ID of the guild to fetch.

        Returns
        -------
        A Guild object representing the requested guild.

        Raises
        ------
        HTTPStatusError: If the API request fails.
        """
        res = await self._http.client.request(method="GET", url=f"/guilds/{guild_id}")
        res.raise_for_status()
        return Guild(**res.json())

    def add_view(self, view: View) -> None:
        """
        Registers a :class:`~dismake.ui.View` for persistent listening.

        Parameters
        ----------
        view: :class:`View`
            The view object to register for dispatching.
        """
        components = list()
        for row in view.rows:
            for component in row.components:
                components.append(component)

        if not components:
            return
        for component in components:
            if self._components.get((custom_id := component.custom_id)):
                continue
            self._components[custom_id] = component

    def add_modal(self, modal: Modal) -> None:
        """
        Registers a :class:`dismake.ui.Modal` for presistent listening.

        Parameters
        ----------
        modal: :class:`Modal`
            The modal object to register for dispatching.
        """
        self._modals[modal.custom_id] = modal

    def add_command(self, command: Union[Command, Group]):
        """
        The add_command function is used to add a command or group of commands to the application.

        Parameters
        ----------
        command: Union[Command, Group]
            The command you want to add to dismake.Bot
        """
        self._commands[command.name] = command

    def command(
        self,
        name: str | None = None,
        description: str = "No description provided.",
        *,
        guild_id: int | None = None,
        default_member_permissions: Permissions | None = None,
        guild_only: bool | None = None,
        nsfw: bool | None = None,
        name_localizations: dict[str, str] | None = None,
        description_localizations: dict[str, str] | None = None,
    ):
        """
        The `command` function is a decorator that registers a function as an application command.

        Parameters
        ----------
        name : str
            The name of the command.
        description : str
            The description of the command.
        guild_id : int | None
            The guild ID, if you want this command to only be visible on a specific guild.
        default_member_permissions : Permissions | None
            The permissions a user needs to invoke this command.
        guild_only : bool | None
            If set to True, this command will only be visible to guilds, not in user DM channels.
        nsfw : bool | None
            If set to True, this command will only be visible in NSFW channels.
        name_localizations : dict[str, str] | None
            Localization dictionary for name field. Values follow the same restrictions as name
        description_localizations : dict[str, str] | None
            Localization dictionary for description field. Values follow the same restrictions as description
        """

        def decorator(coro: AsyncFunction):
            @wraps(coro)
            def wrapper(*_, **__):
                command = Command(
                    name=name if name else coro.__name__,
                    description=description,
                    guild_id=guild_id,
                    callback=coro,
                    nsfw=nsfw,
                    default_member_permissions=default_member_permissions,
                    guild_only=guild_only,
                    name_localizations=name_localizations,
                    description_localizations=description_localizations,
                )
                self._commands[command.name] = command

            return wrapper()

        return decorator

    def create_group(
        self,
        name: str,
        description: str,
        guild_id: int | None = None,
        name_localizations: dict[str, str] | None = None,
        description_localizations: dict[str, str] | None = None,
        default_member_permissions: Permissions | None = None,
        guild_only: bool | None = None,
        nsfw: bool | None = None,
    ):
        """
        The create_group function is a helper function that creates a new Group object and adds it to the list of commands.

        Parameters
        ----------
        name: str
            The name of the command.
        description: str
            The description of the command.
        guild_id : int | None
            The guild ID, if you want this command to only be visible on a specific guild.
        default_member_permissions : Permissions | None
            The permissions a user needs to invoke this command.
        guild_only : bool | None
            If set to True, this command will only be visible to guilds, not in user DM channels.
        nsfw : bool | None
            If set to True, this command will only be visible in NSFW channels.
        name_localizations : dict[str, str] | None
            Localization dictionary for name field. Values follow the same restrictions as name
        description_localizations : dict[str, str] | None
            Localization dictionary for description field. Values follow the same restrictions as description
        """
        command = Group(
            name=name,
            description=description,
            guild_id=guild_id,
            name_localizations=name_localizations,
            description_localizations=description_localizations,
            default_member_permissions=default_member_permissions,
            guild_only=guild_only,
            nsfw=nsfw,
        )
        self._commands[command.name] = command
        return command



from __future__ import annotations
from typing import List, Literal, Optional
from datetime import datetime
from pydantic import BaseModel
from contextlib import suppress


class EmbedField(BaseModel):
    name: str
    value: str
    inline: bool


class EmbedFooter(BaseModel):
    text: str
    icon_url: Optional[str]
    proxy_icon_url: Optional[str]


class EmbedAsset(BaseModel):
    url: str
    proxy_url: Optional[str]
    height: Optional[int]
    width: Optional[int]


class EmbedProvider(BaseModel):
    name: Optional[str]
    url: Optional[str]


class EmbedAuthor(BaseModel):
    name: str
    url: Optional[str]
    icon_url: Optional[str]
    proxy_icon_url: Optional[str]


class Embed(BaseModel):
    title: Optional[str] = None
    type: Literal["rich", "image", "video", "gifv", "article", "link"] = "rich"
    description: Optional[str] = None
    url: Optional[str] = None
    timestamp: Optional[datetime] = None
    color: Optional[int] = None
    footer: Optional[EmbedFooter] = None
    fields: List[EmbedField] = []
    image: Optional[EmbedAsset] = None
    thumbnail: Optional[EmbedAsset] = None
    video: Optional[EmbedAsset] = None
    provider: Optional[EmbedProvider] = None
    author: Optional[EmbedAuthor] = None

    def add_field(self, name: str, value: str, inline: bool = True):
        self.fields.append(EmbedField(name=name, value=value, inline=inline))
        return self

    def get_field(self, index: int) -> Optional[EmbedField]:
        try:
            field = self.fields[index]
        except IndexError:
            return None
        else:
            return field

    def set_footer(
        self, text: str, icon_url: Optional[str], proxy_icon_url: Optional[str]
    ):
        footer = {"text": text}
        if icon_url:
            footer["icon_url"] = icon_url
        if proxy_icon_url:
            footer["proxy_icon_url"] = proxy_icon_url
        self.footer = EmbedFooter(**footer)
        return self

    def set_image(
        self,
        url: str,
        proxy_url: Optional[str],
        width: Optional[int],
        height: Optional[int],
    ):
        self.image = EmbedAsset(
            url=url, proxy_url=proxy_url, width=width, height=height
        )
        return self

    def set_provider(self, name: str, url: str):
        self.provider = EmbedProvider(name=name, url=url)
        return self

    def set_author(
        self,
        name: str,
        url: Optional[str],
        icon_url: Optional[str],
        proxy_icon_url: Optional[str],
    ):
        self.author = EmbedAuthor(
            name=name, url=url, icon_url=icon_url, proxy_icon_url=proxy_icon_url
        )
        return self



from __future__ import annotations

from datetime import datetime
from typing import Optional, Union
from pydantic import BaseModel, PrivateAttr
from fastapi import Request

from .user import User
from ..types import SnowFlake
from .permission_overwrites import PermissionOverwrites

__all__ = (
    "PartialMessagable",
    "TextChannel",
    "CategoryChannel",
    "Channel",
    "AnnouncementChannel",
)


class PartialMessagable(BaseModel):
    _request: Request = PrivateAttr()
    id: SnowFlake
    type: int

    @property
    def mention(self) -> str:
        return f"<#{self.id}>"

    def __repr__(self) -> str:
        return f"<Channel id={self.id}>"

    class Config:
        arbitrary_types_allowed = True


class TextChannel(PartialMessagable):
    guild_id: Optional[int]
    position: Optional[int]
    permission_overwrites: list[PermissionOverwrites] = list()
    name: Optional[str]
    topic: Optional[str]
    nsfw: Optional[bool] = False
    last_message_id: Optional[int]
    parent_id: Optional[int]
    last_pin_timestamp: Optional[datetime]
    permissions: Optional[str]
    rate_limit_per_user: Optional[int]
    default_auto_archive_duration: Optional[int]
    created_at: Optional[datetime]

    def __repr__(self) -> str:
        return f"<TextChannel name={self.name}>"

    def __str__(self) -> str:
        return f"{self.name}"


class CategoryChannel(PartialMessagable):
    permission_overwrites: list[PermissionOverwrites] = list()
    name: Optional[str]
    nsfw: Optional[bool] = False
    position: int
    guild_id: SnowFlake

    def __repr__(self) -> str:
        return f"<CategoryChannel name={self.name}>"

    def __str__(self) -> str:
        return str(self.name)


class AnnouncementChannel(PartialMessagable):
    guild_id: SnowFlake
    name: Optional[str]
    type: int
    position: Optional[int]
    permission_overwrites: list[PermissionOverwrites] = list()
    nsfw: Optional[bool] = False
    topic: Optional[str]
    last_message_id: Optional[SnowFlake]
    parent_id: Optional[SnowFlake]
    default_auto_archive_duration: Optional[int]


class DMChannel(PartialMessagable):
    last_message_id: Optional[SnowFlake]
    recipitents: Optional[list[User]]
    icon: Optional[str]


class GroupDmChannel(DMChannel):
    owner_id: Optional[SnowFlake]


Channel = Union[
    TextChannel, CategoryChannel, AnnouncementChannel, DMChannel, GroupDmChannel
]



from __future__ import annotations
from typing import Any, Optional

from pydantic import BaseModel
from ..types import SnowFlake

__all__ = ("Role",)


class RoleTags(BaseModel):
    bot_id: Optional[SnowFlake]
    integration_id: Optional[SnowFlake]
    premium_subscriber: Optional[Any]
    subscription_listing_id: Optional[SnowFlake]
    available_for_purchase: Optional[Any]
    guild_connections: Optional[Any]


class Role(BaseModel):
    id: SnowFlake
    name: str
    color: int
    hoist: bool
    icon: Optional[str]
    unicode_emoji: Optional[str]
    position: int
    permissions: int
    managed: bool
    mentionable: bool
    tags: Optional[RoleTags]

    def __str__(self) -> str:
        return self.name

    @property
    def mention(self) -> str:
        return f"<@&{self.id}>"



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from ..enums import ComponentType
from ..models import PartialEmoji
from .component import Component

if TYPE_CHECKING:
    from ..models import Interaction
    from typing_extensions import Self

__all__ = (
    "SelectOption",
    "StringSelectMenu",
)


class BaseSelect(Component):
    """
    Represents a select menu.

    This class should not be initialized directly.
    All other select menus should subclass this class.

    Parameters
    ----------
    type: :class:`ComponentType`
        The type of the component.
    custom_id: :class:`str`
        The custom ID of the component. If not provided, a random UUID will be generated.
    placeholder: :class:`str`
        The text displayed on the select menu.
    disabled: :str:`bool`
        Indicates whether the component is disabled.
    min_values: :class:`int`
        Determines the minimum number of options that can be selected by the user. (default: 1)
    mxn_values: :class:`int`
        Determines the maximum number of options that can be selected by the user. (default: 1)

    Attributes
    ----------
    view: :class:`View`
        The view associated with the component.
    """

    def __init__(
        self,
        type: ComponentType,
        custom_id: str | None = None,
        placeholder: str | None = None,
        disabled: bool = False,
        min_values: int = 1,
        max_values: int = 1,
    ) -> None:
        super().__init__(type=type, custom_id=custom_id, disabled=disabled)
        self.placeholder = placeholder
        self.min_values = min_values
        self.max_values = max_values

    def to_dict(self) -> dict[str, Any]:
        base = super().to_dict()
        if self.placeholder is not None:
            base["placeholder"] = self.placeholder
        base.update({"min_value": self.min_values, "max_values": self.max_values})
        if self.disabled is not None:
            base["disabled"] = self.disabled
        return base


class SelectOption:
    """
    Represents a select menu option.

    Parameters
    ----------
    label: :class:`str`
        The text displayed on the option.
    value: :class:`str`
        The option's value.
    description: :class:`str`
        The description displayed on the option.
    emoji: :class:`PartialEmoji`
        The emoji displayed on the option
    default: :class:`bool`
        Indicates whether the option is default.
    """

    __slots__ = ("label", "value", "description", "default", "emoji")

    def __init__(
        self,
        label: str,
        value: str | None = None,
        description: str | None = None,
        emoji: PartialEmoji | None = None,
        default: bool | None = None,
    ) -> None:
        self.label = label
        self.value = value or label
        self.description = description
        self.default = default
        self.emoji: PartialEmoji | None
        if emoji is not None:
            if isinstance(emoji, str):
                self.emoji = PartialEmoji.from_str(emoji)
            else:
                self.emoji = emoji
        else:
            self.emoji = None

    def to_dict(self) -> dict[str, Any]:
        """
        Converts a select option into a dict.

        Returns
        -------
        dict[str, Any]
        """
        base: dict[str, Any] = {
            "label": self.label,
            "value": self.value,
        }
        if self.description is not None:
            base["description"] = self.description
        if self.default is not None:
            base["default"] = self.default
        if self.emoji is not None:
            base["emoji"] = self.emoji.dict(exclude_none=True)
        return base


class StringSelectMenu(BaseSelect):
    """
    Represents a String select menu.

    Parameters
    ----------
    options: list[:class:`SelectOption`]
        The list of options that will be shown when the user clicks on the menu.
    custom_id: :class:`str`
        The custom ID of the component. If not provided, a random UUID will be generated.
    placeholder: :class:`str`
        The text displayed on the select menu.
    disabled: :str:`bool`
        Indicates whether the component is disabled.
    min_values: :class:`int`
        Determines the minimum number of options that can be selected by the user. (default: 1)
    mxn_values: :class:`int`
        Determines the maximum number of options that can be selected by the user. (default: 1)

    Attributes
    ----------
    view: :class:`View`
        The view associated with the component.
    """

    def __init__(
        self,
        options: list[SelectOption],
        custom_id: str | None = None,
        placeholder: str | None = None,
        disabled: bool = False,
        min_values: int = 1,
        max_values: int = 1,
    ):
        super().__init__(
            type=ComponentType.STRING_SELECT,
            custom_id=custom_id,
            placeholder=placeholder,
            min_values=min_values,
            max_values=max_values,
            disabled=disabled,
        )
        self.options = options

    def add_option(self, option: SelectOption) -> Self:
        """
        Append a option to this select menu

        Returns
        -------
        Self
        """
        self.options.append(option)
        return self

    def to_dict(self) -> dict[str, Any]:
        """
        Converts a :class:`StringSelectMenu` into a dict.

        Returns
        -------
        dict[str, Any]
        """
        base = super().to_dict()
        base["options"] = [o.to_dict() for o in self.options]
        return base



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Optional
from functools import wraps
from .commands import Command, Group
from .errors import PluginException

if TYPE_CHECKING:
    from .commands import Command, Group
    from .client import Bot
    from .types import AsyncFunction
    from .permissions import Permissions

__all__ = ("Plugin",)


class Plugin:
    """
    Represents a plugin.

    Plugins help you split your code into different files,
    similar to `APIRouter` and discord.py's `Cog`.

    Parameters
    ----------
    name: :class:`str`
        The name of the plugin.
    default_member_permissions: :class:`Permissions`
        This will override all the commands permissions with this permission.

    Attributes
    ----------
    bot: :class:`Bot`
        The bot instance.
    """

    def __init__(
        self,
        name: str = __name__,
        default_member_permissions: Permissions | None = None,
    ) -> None:
        self.name = name
        self.bot: Bot
        self._commands: dict[str, Command | Group] = {}
        self.error_handler: Optional[AsyncFunction] = None
        self._on_load: AsyncFunction | None = None
        self._events: dict[str, list[AsyncFunction]] = {}
        self.default_member_permissions = default_member_permissions

    def on_load(self, coro: AsyncFunction):
        @wraps(coro)
        def wrapper(*_, **__):
            if not asyncio.iscoroutinefunction(coro):
                raise PluginException(f"{coro.__name__!r} is not coroutine function.")
            self._on_load = coro
            return coro

        return wrapper()

    def event(self, event_name: str | None = None):
        """
        A decorator that registers an event to listen to.

        Parameters
        ----------
        event_name: :class:`str`
            The event name you want to listen.

        Example usage
        -------------
            import dismake

            plugin = dismake.Plugin()

            @plugin.event()
            async def on_interaction():
                print(f"A new interaction received.")
        """

        def decorator(coro: AsyncFunction):
            @wraps(coro)
            def wrapper(*_, **__):
                if not asyncio.iscoroutinefunction(coro):
                    raise PluginException(
                        f"{coro.__name__!r} is not coroutine function."
                    )
                name = event_name or coro.__name__
                if self._events.get(name) is not None:
                    self._events[name].append(coro)
                else:
                    self._events[name] = [coro]

            return wrapper()

        return decorator

    def command(
        self,
        name: str,
        description: str = "No description provided,",
        *,
        guild_id: int | None = None,
        default_member_permissions: Permissions | None = None,
        guild_only: bool | None = None,
        nsfw: bool | None = None,
        name_localizations: dict[str, str] | None = None,
        description_localizations: dict[str, str] | None = None,
        plugin_permissions: bool = True,
    ):
        """
        The `command` function is a decorator that registers a function as an application command.

        Parameters
        ----------
        name: :class:`str`
            The name of the command.
        description: :class:`str`
            The description of the command.
        guild_id: :class:`int | None`
            The guild ID, if you want this command to only be visible on a specific guild.
        default_member_permissions: :class:`Permissions | None`
            The permissions a user needs to invoke this command.
        guild_only: :class:`bool | None`
            If set to True, this command will only be visible to guilds, not in user DM channels.
        nsfw: :class:`bool | None`
            If set to True, this command will only be visible in NSFW channels.
        name_localizations: :class:`dict[str, str] | None`
            Localization dictionary for name field. Values follow the same restrictions as name.
        description_localizations: :class:`dict[str, str] | None`
            Localization dictionary for description field. Values follow the same restrictions as description.
        plugin_permissions: :class:`bool`
            If this set to false then the plugin won't override permissions for this command.
        """

        def decorator(coro: AsyncFunction):
            @wraps(coro)
            def wrapper(*_, **__):
                nonlocal default_member_permissions
                if not asyncio.iscoroutinefunction(coro):
                    raise PluginException(
                        f"{coro.__name__!r} command callback must be a coroutine function."
                    )

                if self.default_member_permissions is not None and plugin_permissions:
                    default_member_permissions = self.default_member_permissions

                command = Command(
                    name=name,
                    description=description,
                    guild_id=guild_id,
                    callback=coro,
                    nsfw=nsfw,
                    default_member_permissions=default_member_permissions,
                    guild_only=guild_only,
                    name_localizations=name_localizations,
                    description_localizations=description_localizations,
                )
                command.plugin = self
                self._commands[command.name] = command
                return command

            return wrapper()

        return decorator

    def create_group(
        self,
        name: str,
        description: str,
        guild_id: int | None = None,
        name_localizations: dict[str, str] | None = None,
        description_localizations: dict[str, str] | None = None,
        default_member_permissions: Permissions | None = None,
        guild_only: bool | None = None,
        nsfw: bool | None = None,
        parent: Group | None = None,
        plugin_permissions: bool = True,
    ):
        """
        The create_group function is a helper function that creates a new Group object and adds it to the list of commands.

        Parameters
        ----------
        name: :class:`str`
            The name of the command.
        description: :class:`str`
            The description of the command.
        guild_id: :class:`int | None`
            The guild ID, if you want this command to only be visible on a specific guild.
        default_member_permissions : :class:`Permissions | None`
            The permissions a user needs to invoke this command.
        guild_only: :class:`bool | None`
            If set to True, this command will only be visible to guilds, not in user DM channels.
        nsfw: :class:`bool | None`
            If set to True, this command will only be visible in NSFW channels.
        name_localizations: :class:`dict[str, str] | None`
            Localization dictionary for name field. Values follow the same restrictions as name.
        description_localizations: :class:`dict[str, str] | None`
            Localization dictionary for description field. Values follow the same restrictions as description.
        plugin_permissions: :class:`bool`
            If this set to false then the plugin won't override permissions for this command.
        """
        if self.default_member_permissions is not None and plugin_permissions:
            default_member_permissions = self.default_member_permissions

        command = Group(
            name=name,
            description=description,
            guild_id=guild_id,
            name_localizations=name_localizations,
            description_localizations=description_localizations,
            default_member_permissions=default_member_permissions,
            guild_only=guild_only,
            nsfw=nsfw,
            parent=parent,
        )
        command.plugin = self
        self._commands[command.name] = command
        return command

    async def load(self, bot: Bot):
        """
        Loads the plugin.

        Parameters
        ----------
        bot: :class:`Bot`
            The bot instance.
        """
        bot._commands.update(self._commands)
        bot._events.update(self._events)
        if self._on_load is not None:
            asyncio.create_task(self._on_load())
        self.bot = bot



from pydantic import BaseModel
from typing import Dict, List, Optional, Union
from ..types import SnowFlake


__all__ = (
    "AppCommandChoice",
    "AppCommandOption",
    "AppCommand",
)


class AppCommandChoice(BaseModel):
    name: str
    value: Union[str, int, float]


class AppCommandOption(BaseModel):
    type: int
    name: str
    name_localizations: Optional[Dict[str, str]]
    description: Optional[str] = "No description provided"
    description_localizations: Optional[Dict[str, str]]
    required: Optional[bool] = False
    choices: Optional[List[AppCommandChoice]]
    options: Optional[List["AppCommandOption"]]
    channel_types: Optional[List[str]]
    min_value: Optional[int]
    max_value: Optional[int]
    min_length: Optional[int]
    max_length: Optional[int]
    autocomplete: Optional[bool]


class AppCommand(BaseModel):
    id: SnowFlake
    type: int = 1
    App_id: SnowFlake
    guild_id: Optional[SnowFlake]
    name: str
    name_localizations: Optional[Dict[str, str]]
    description: Optional[str] = "No description provided"
    description_localizations: Optional[Dict[str, str]]
    options: Optional[List[AppCommandOption]]
    default_member_permissions: Optional[str]
    dm_permission: Optional[bool] = True
    nsfw: Optional[bool] = False
    version: int



from .application_command import *
from .user import *
from .embed import *
from .guild import *
from .message import *
from .role import *
from .interaction import *
from .emoji import *
from .components import *
from .permission_overwrites import *
from .channels import *



from __future__ import annotations

from typing import Any, List, Optional, TYPE_CHECKING
from functools import wraps

from ..enums import ButtonStyles, ComponentType
from ..types import AsyncFunction
from .component import Component
from .button import Button
from .select import StringSelectMenu


if TYPE_CHECKING:
    from ..models import Interaction
    from .select import SelectOption
    from typing_extensions import Self
__all__ = ("View",)


class Row:
    """
    Represents an Action Row.
    """

    def __init__(self):
        self.components: List[Component] = list()

    @property
    def is_full(self) -> bool:
        if not self.components:
            return False
        if isinstance(self.components[0], (StringSelectMenu)):
            return len(self.components) == 1
        return len(self.components) == 5

    def add_component(self, component: Component):
        if self.is_full:
            raise ValueError("can't able to find free space to add the component.")
        self.components.append(component)
        return self

    def to_dict(self) -> dict[str, Any]:
        return {
            "type": ComponentType.ACTION_ROW.value,
            "components": [i.to_dict() for i in self.components],
        }


class View:
    """
    Represents a view that contains message components.

    Example
    --------
        view = View()
        @view.buttoon(label="Click me")
        async def click_me(ctx):
            await ctx.respond("Clicked")
    """

    def __init__(self) -> None:
        self.rows: List[Row] = list()
        self._error_handler: AsyncFunction = self.on_error

    @property
    def is_full(self) -> bool:
        return len(self.rows) == 5 and all(row.is_full for row in self.rows)

    async def on_error(self, interaction: Interaction, e: Exception) -> Any:
        pass

    def add_component(self, component: Component) -> Self:
        if self.is_full:
            raise ValueError("can't able to find free space to add the component.")

        if isinstance(component, Button):
            if not self.rows or self.rows[-1].is_full:
                self.rows.append(Row())
            self.rows[-1].add_component(component)
        else:
            self.rows.append(Row().add_component(component))
        return self

    def button(
        self,
        label: Optional[str] = None,
        custom_id: Optional[str] = None,
        emoji: Optional[str] = None,
        style: Optional[ButtonStyles] = None,
        url: Optional[str] = None,
        disabled: Optional[bool] = None,
    ):
        def decorator(coro: AsyncFunction):
            @wraps(coro)
            def wrapper(*_, **__):
                button = Button(
                    label=label,
                    custom_id=custom_id,
                    style=style,
                    url=url,
                    disabled=disabled,
                    emoji=emoji,
                )
                button._callback = coro
                self.add_component(button)
                return button

            return wrapper()

        return decorator

    def string_select(
        self,
        options: List[SelectOption],
        placeholder: Optional[str] = None,
        custom_id: Optional[str] = None,
        min_values: int = 1,
        max_values: int = 1,
        disabled: bool = False,
    ):
        def decorator(coro: AsyncFunction):
            @wraps(coro)
            def wrapper(*_, **__):
                select = StringSelectMenu(
                    placeholder=placeholder,
                    options=options,
                    custom_id=custom_id,
                    min_values=min_values,
                    max_values=max_values,
                    disabled=disabled,
                )
                select._callback = coro
                self.add_component(select)
                return select

            return wrapper()

        return decorator

    def add_url_button(
        self, label: str, url: str, emoji: str, disabled: Optional[bool]
    ):
        button = Button(
            label=label,
            url=url,
            style=ButtonStyles.link,
            emoji=emoji,
            disabled=disabled,
            custom_id=None,
        )
        self.add_component(button)
        return button

    def error(self, coro: AsyncFunction):
        @wraps(coro)
        def wrapper(*_, **__):
            self._error_handler = coro
            return coro

        return wrapper()

    def to_dict(self) -> list[dict[str, Any]]:
        return [row.to_dict() for row in self.rows]



from __future__ import annotations
from typing import List, Optional, Union
from httpx import AsyncClient, Response

from .models import User
from .models import AppCommand
from .commands import Command, Group

__all__ = ("HttpClient",)


class HttpClient:
    def __init__(
        self,
        *,
        token: str,
        client_id: int,
    ) -> None:
        self.token = token
        self.client_id = client_id
        self.api_version = 10
        self.app_command_endpoint = f"/applications/{client_id}/commands"
        self.client = AsyncClient(base_url=self.base_url, headers=self.headers)
        self._user: User

    @property
    def base_url(self) -> str:
        return "https://discord.com/api/v%s/" % self.api_version

    @property
    def headers(self) -> dict:
        return {"Authorization": "Bot %s" % self.token}

    # async def register_command(self, command: SlashCommand):
    #     return await self.client.request(
    #         method="POST",
    #         url="/applications/%s/commands" % self.client_id,
    #         json=command.to_dict(),
    #         headers=self.headers,
    #     )

    async def get_global_commands(self) -> list[AppCommand] | None:
        res = await self.client.request(
            method="GET",
            url=f"/applications/{self.client_id}/commands",
            headers=self.headers,
        )
        res.raise_for_status()
        _json = res.json()
        if not _json:
            return list()

        return [AppCommand(**command) for command in _json]

    async def bulk_override_commands(
        self, commands: List[Union[Command, Group]], guild_id: Optional[int] = None
    ) -> Response:
        return await self.client.request(
            method="PUT",
            url=f"/applications/{self.client_id}/commands",
            json=[command.to_dict() for command in commands],
            headers=self.headers,
        )

    async def remove_all_commands(self):
        return await self.client.request(
            method="PUT",
            url=f"/applications/{self.client_id}/commands",
            headers=self.headers,
            json=[],
        )

    async def fetch_me(self):
        res = await self.client.request(
            method="GET", url="/users/@me", headers=self.headers
        )
        res.raise_for_status()
        self._user = User(**res.json())



from dismake import ui
import dismake
from dismake.models import Interaction
from dismake.models.interaction import Interaction
from dismake.ui import select

view = ui.View()


@view.string_select(
    options=[
        ui.SelectOption(label="Lund"),
        ui.SelectOption(label="Chut"),
        ui.SelectOption(label="Gand"),
    ]
)
async def select_callback(interaction: dismake.Interaction):
    await interaction.edit_message(f"You want {interaction.data.values[0]}")


@view.button(label="1")
async def b1(interaction: dismake.Interaction):
    await interaction.send("Hmm What ?", ephemeral=True)


@view.button(label="2")
async def b2(interaction: dismake.Interaction):
    await interaction.send("Hmm What ?", ephemeral=True)


# Modal


class MyModal(ui.Modal):
    def __init__(self) -> None:
        super().__init__("My modal")
        self.add_item(ui.TextInput("Name"))
        self.add_item(ui.TextInput("Age"))

    async def on_submit(self, interaction: Interaction):
        return await interaction.send(
            f"Name: {self.children[0].value}\nAge: {self.children[1].value}"
        )



import dismake
import config
from plugins import misc


app = dismake.Bot(
    token=config.TOKEN, client_id=config.CLIENT_ID, client_public_key=config.PUBLIC_KEY
)


@app.event()
async def on_ready():
    """This event is called when the bot is ready to recive interactions from discord."""
    misc.plugin.load(app)
    print("Logged in as %s" % app.user)

    # Sync all the slash commands to discord
    await app.sync_commands()


if __name__ == "__main__":
    app.run(app="main:app", reload=True)

# Question: What does 'main:app' represent?
# Answer: 'main:app' represents the name of the main module
# and the variable or object within that module that will be used as the dismake.Bot.

# Question: Why is 'reload=True' included?
# Answer: The 'reload=True' parameter enables automatic reloading of the
# uvicorn when changes are made to the code.
# This allows you to see the updates without manually restarting the server.



from __future__ import annotations

from typing import Any, List, Union, Optional
from datetime import datetime
from pydantic import BaseModel, PrivateAttr
from fastapi import Request
from .embed import Embed

from ..types import SnowFlake
from .user import Member, User

__all__ = ("Message",)


class MessageInteraction(BaseModel):
    id: SnowFlake
    type: int
    name: str
    user: User
    member: Optional[Member]


class Message(BaseModel):
    _request: Request = PrivateAttr()
    type: int
    id: SnowFlake
    channel_id: SnowFlake
    author: User
    content: str
    timestamp: datetime
    tts: bool
    mention_everyone: bool
    mentions: List[User]
    mention_roles: List[int]
    pinned: bool
    edited_timestamp: Optional[datetime]
    mention_channels: Optional[List[Any]]
    attachments: Optional[List[Any]]
    embeds: Optional[List[Embed]]
    reactions: Optional[List[Any]]
    nonce: Optional[Union[str, int]]
    webhook_id: Optional[SnowFlake]
    activity: Optional[Any]
    application: Optional[Any]
    application_id: Optional[Any]
    message_reference: Optional[Any]
    flags: Optional[Any]
    referenced_message: Optional[Any]
    interaction: Optional[MessageInteraction]
    thread: Optional[Any]
    components: Optional[List[Any]]
    sticker_items: Optional[List[Any]]
    stickers: Optional[List[Any]]
    position: Optional[int]
    role_subscription_data: Optional[Any]



from __future__ import annotations

import inspect
from typing import Any, TYPE_CHECKING, Optional
from functools import wraps

from .permissions import Permissions
from .types import AsyncFunction
from .models import (
    User,
    Member,
    Role,
    Channel,
    TextChannel,
    CategoryChannel,
    AnnouncementChannel,
    ApplicationCommandData,
)
from .errors import CommandInvokeError
from .enums import ChannelType, CommandType, OptionType, Locale

if TYPE_CHECKING:
    from .types import AsyncFunction
    from .permissions import Permissions
    from .plugin import Plugin
    from .models import Interaction


__all__ = ("Command", "Option", "Choice", "Group", "Argument")


_option_types = {
    # fmt: off
    bool:           OptionType.BOOLEAN,
    int:            OptionType.INTEGER,
    str:            OptionType.STRING,
    User:           OptionType.USER,
    Member:         OptionType.USER,
    Role:           OptionType.ROLE,
    # Channels
    Channel:        OptionType.CHANNEL,
    TextChannel:        OptionType.CHANNEL,
    CategoryChannel:        OptionType.CHANNEL,
    AnnouncementChannel:        OptionType.CHANNEL,
}


def _get_options(func: AsyncFunction) -> tuple[Option]:
    """
    Extracts options from the given command callback.

    Parameters
    ----------
    func : AsyncFunction
        The function to get the options from.

    Returns
    -------
    tuple[Option]
        The options of the command.
    """
    params = inspect.signature(func).parameters
    options: tuple[Option] = tuple()
    for k, v in params.items():
        # k: The name of the function
        #     - name
        # v: The annotation of the function
        #     - Argument[str, <Option name="foo">]
        annotation = v.annotation
        if isinstance(annotation, tuple):
            option_type: type = annotation[0]
            option_object: Option = annotation[1]
            option_object.type = _option_types[option_type]
            if option_object.description is None:
                option_object.description = "..."

            if option_object.name is None:
                option_object.name = k

            if option_object.required is None:
                if v.default != inspect._empty:
                    option_object.required = False
                else:
                    option_object.required = True

            options += (option_object,)
    return options

def _populate_locales(locale: dict[Locale, str]) -> dict[str, str]:
    locales = dict()
    for lang, value in locale.items():
        locales[lang.value] = value
    return locales

if TYPE_CHECKING:
    from typing_extensions import Annotated as Argument
else:
    class Argument:
        """
        This class used to annotate arguments of a command.
        """
        def __class_getitem__(cls, obj: type) -> type:
            if not isinstance(obj, tuple):
                raise TypeError(f"Argument must be a tuple, not {type(obj)}")
            if not isinstance(obj[1], Option):
                raise TypeError(
                    f"Second option of an 'Argument' must be a dismake.Option, not {type(obj[1])}"
                )
            return obj

class Command:
    """
    Represents a Discord command or a Discord sub-command.

    Parameters
    ----------
    name (str):
        The name of the command/sub-command.
    description (str):
        A brief description of what the command/sub-command does.(Max length 100)
    callback (AsyncFunction):
        The function to be executed when the command is called.
    guild_id (int|None):
        The ID of the guild this command is registered in, or None if it's a global command.
    name_localizations (dict[str, str]|None):
        A dictionary of localized names for the command, keyed by language code.
    description_localizations (dict[str, str]|None):
        A dictionary of localized descriptions for the command, keyed by language code.
    default_member_permissions (Permissions|None):
        The default permissions required for members to execute the command.
    guild_only (bool|None):
        Whether the command can only be executed in a guild or not.
    dm_permission (bool):
        Whether the command can be executed in DMs or not.
    nsfw (bool|None):
        Whether the command can only be executed in channels marked as NSFW or not.
    """

    def __init__(
        self,
        name: str,
        description: str,
        callback: AsyncFunction,
        guild_id: int | None = None,
        name_localizations: dict[str, str] | None = None,
        description_localizations: dict[str, str] | None = None,
        default_member_permissions: Permissions | None = None,
        guild_only: bool | None = None,
        nsfw: bool | None = None,
    ) -> None:
        self.name = name
        self.description = description
        self.callback = callback
        self.guild_id = guild_id
        self.name_localizations = name_localizations
        self.description_localizations = description_localizations
        self.default_member_permissions = default_member_permissions
        self.dm_permission = not guild_only
        self.nsfw = nsfw
        self.parent: Group | None = None
        self.type: CommandType | OptionType = (
            CommandType.SLASH if self.parent is not None else OptionType.SUB_COMMAND
        )
        self.options = _get_options(self.callback)
        self.plugin: Plugin | None = None
        self.autocompletes: dict[str, AsyncFunction] = {}
        self.error_handler: Optional[AsyncFunction] = None

    def __str__(self) -> str:
        return self.name

    async def _invoke_error_handlers(
        self, interaction: Interaction, error: CommandInvokeError
    ):
        """
        Invokes the error handlers for the command.

        Parameters
        ----------
        interaction: Interaction
            The interaction that triggered the error.
        error: CommandInvokeError
            The error that triggered the command.
        """
        if self.error_handler is not None:
            return await self.error_handler(interaction, error)
        if self.plugin is not None and self.plugin.error_handler is not None:
            return await self.plugin.error_handler(interaction, error)
        if (bot_error_handler := interaction.bot.error_handler) is not None:
            return await bot_error_handler(interaction, error)
        raise error

    async def invoke(self, interaction: Interaction):
        """
        Invokes the command.

        Parameters
        ----------
        interaction: Interaction
            The interaction that triggered the command.

        Raises
        ------
        CommandInvokeError
            The command failed to invoke.
        """
        args: tuple = tuple()
        kwargs: dict = dict()
        options = interaction.namespace.__dict__
        params = inspect.signature(self.callback).parameters
        for k, v in params.items():
            option: type | None = options.get(k)
            if option is not None:
                if v.default != inspect._empty:
                    kwargs[k] = option
                else:
                    args += (option,)
        try:
            await self.callback(interaction, *args, **kwargs)
        except Exception as e:
            assert interaction.data is not None and isinstance(
                interaction.data, ApplicationCommandData
            )
            exception = CommandInvokeError(self, e)
            return await self._invoke_error_handlers(interaction, exception)

    async def invoke_autocomplete(self, interaction: Interaction, name: str):
        """
        Invokes the autocomplete for the command.

        Parameters
        ----------
        interaction: Interaction
            The interaction that triggered the command.
        name: str
            The name of the option to autocomplete.

        Raises
        ------
        CommandInvokeError
            The command failed to invoke.
        """
        autocomplete = self.autocompletes.get(name)
        if not autocomplete:
            return

        choices: list[Choice] | None = await autocomplete(
            interaction, name=interaction.namespace.__dict__[name]
        )
        if choices is not None:
            return await interaction.autocomplete(choices)

    def autocomplete(self, option: str):
        """
        Decorator that registers an autocomplete for the command.

        Parameters
        ----------
        option: str
            The name of the option to autocomplete.

        Returns
        -------
        Callable[[AsyncFunction], AsyncFunction]
            The decorator.
        """

        def decorator(coro: AsyncFunction):
            @wraps(coro)
            def wrapper(*_, **__):
                self.autocompletes[option] = coro
                return coro

            return wrapper()

        return decorator

    def to_dict(self) -> dict[str, Any]:
        """
        Converts the command into a dictionary.

        Returns
        -------
        dict[str, Any]
            The dictionary.
        """
        base = {
            "name": self.name,
            "description": self.description,
            "type": self.type.value,
            "options": [option.to_dict() for option in self.options]
            if self.options
            else [],
        }
        if self.parent is None:
            if self.guild_id is not None:
                base["guild_id"] = self.guild_id
            if self.name_localizations is not None:
                base["name_localizations"] = self.name_localizations
            if self.description_localizations is not None:
                base["description_localizations"] = self.description_localizations
            if self.default_member_permissions is not None:
                base[
                    "default_member_permissions"
                ] = self.default_member_permissions.value
            if self.dm_permission is not None:
                base["dm_permissions"] = self.dm_permission
            if self.nsfw is not None:
                base["nsfw"] = self.nsfw
        return base


class Group:
    """
    Represents a slash command group.

    Attributes
    ----------
    name (str):
        The name of the group.
    description (str):
        A brief description of what the group does.(Max length 100)
    guild_id (int|None):
        The ID of the guild this command is registered in, or None if it's a global command.
    name_localizations (dict[str, str]|None):
        A dictionary of localized names for the group, keyed by language code.
    description_localizations (dict[str, str]|None):
        A dictionary of localized descriptions for the group, keyed by language code.
    default_member_permissions (Permissions|None):
        The default permissions required for members to execute the group.
    guild_only (bool|None):
        Whether the group can only be executed in a guild or not.
    dm_permission (bool):
        Whether the group can be executed in DMs or not.
    nsfw (bool|None):
        Whether the group can only be executed in channels marked as NSFW or not.
    parent (Group|None):
        The parent group, if any.
    """

    def __init__(
        self,
        name: str,
        description: str,
        guild_id: int | None = None,
        name_localizations: dict[str, str] | None = None,
        description_localizations: dict[str, str] | None = None,
        default_member_permissions: Permissions | None = None,
        guild_only: bool | None = None,
        nsfw: bool | None = None,
        parent: Group | None = None,
    ) -> None:
        self.name = name
        self.description = description
        self.guild_id = guild_id
        self.name_localizations = name_localizations
        self.description_localizations = description_localizations
        self.default_member_permissions = default_member_permissions
        self.dm_permission = not guild_only
        self.nsfw = nsfw
        self.parent = parent
        self.type: CommandType | OptionType = (
            CommandType.SLASH if not self.parent else OptionType.SUB_COMMAND_GROUP
        )
        self.commands: dict[str, Command | Group] = {}
        self.plugin: Plugin | None = None
        if self.parent:
            if self.parent.parent:
                raise ValueError("groups can only be nested at most one level")
            self.parent.add_command(self)

        if len(self.commands.values()) > 25:
            raise TypeError("groups cannot have more than 25 commands")

    def __str__(self) -> str:
        return self.name

    def add_command(self, command: Group | Command):
        """
        Adds a command to the group.

        Parameters
        ----------
        command: Group | Command
            The command to add.

        Raises
        ------
        TypeError
            The command is not a group or a command.
        ValueError
            The command is nested more than one level.
        """
        if isinstance(command, Group) and self.parent:
            raise ValueError("groups can only be nested at most one level")

        if not isinstance(command, (Group, Command)):
            raise TypeError("expected Group or Command but got %s" % type(command))
        self.commands[command.name] = command
        command.parent = self
        return command

    def command(
        self,
        name: str | None = None,
        description: str = "No description provided.",
        *,
        guild_id: int | None = None,
        default_member_permissions: Permissions | None = None,
        guild_only: bool | None = None,
        nsfw: bool | None = None,
        name_localizations: dict[str, str] | None = None,
        description_localizations: dict[str, str] | None = None,
    ):
        """
        Decorator that creates a sub command.

        Parameters
        ----------
        name: str | None
            The name of the command.
        description: str
            A brief description of what the command does.(Max length 100)
        guild_id: int | None
            The ID of the guild this command is registered in, or None if it's a global command.
        default_member_permissions: Permissions | None
            The default permissions required for members to execute the command.
        guild_only: bool | None
            Whether the command can only be executed in a guild or not.
        nsfw: bool | None
            Whether the command can only be executed in channels marked as NSFW or not.
        name_localizations: dict[str, str]|None
            A dictionary of localized names for the command, keyed by language code.
        description_localizations: dict[str, str]|None
            A dictionary of localized descriptions for the command, keyed by
        """

        def decorator(coro: AsyncFunction):
            command = Command(
                name=name or coro.__name__,
                description=description,
                guild_id=guild_id,
                callback=coro,
                nsfw=nsfw,
                default_member_permissions=default_member_permissions,
                guild_only=guild_only,
                name_localizations=name_localizations,
                description_localizations=description_localizations,
            )
            self.add_command(command)
            return command

        return decorator

    def create_sub_group(self, name: str, description: str):
        """
        Decorator that creates a sub group.

        Parameters
        ----------
        name: str
            The name of the group.
        description: str
            A brief description of what the group does.(Max length 100)
        """
        command = Group(name=name, description=description, parent=self)
        return command

    def to_dict(self) -> dict[str, Any]:
        """
        Creates a dictionary representation of the group.

        Returns
        -------
        dict[str, Any]
            The dictionary representation of the group.
        """
        base = {
            "name": self.name,
            "description": self.description,
            "type": self.type.value,
            "options": [command.to_dict() for command in self.commands.values()],
        }
        if self.parent is None:
            if self.guild_id is not None:
                base["guild_id"] = self.guild_id
            if self.name_localizations is not None:
                base["name_localizations"] = self.name_localizations
            if self.description_localizations is not None:
                base["description_localizations"] = self.description_localizations
            if self.default_member_permissions is not None:
                base[
                    "default_member_permissions"
                ] = self.default_member_permissions.value
            if self.dm_permission is not None:
                base["dm_permission"] = self.dm_permission
            if self.nsfw is not None:
                base["nsfw"] = self.nsfw
        return base


class Option:
    """
    Represents a slash command option.

    Attributes
    ----------
    name (str):
        The name of the option.
    description (str):
        The description of the option.
    name_localizations (dict[str, str]|None):
        A dictionary of localized names for the option, keyed by language code.
    description_localizations (dict[str, str]|None):
        A dictionary of localized descriptions for the option, keyed by language code.
    required (bool|None):
        Whether the option is required or not.
    choices (list[Choice]|None):
        A list of choices for the option.
    channel_types (list[ChannelType]|None):
        A list of channel types for the option.
    min_value (int|None):
        The minimum value for the option.
    max_value (int|None):
        The maximum value for the option.
    autocomplete (bool|None):
        Whether the option is an autocomplete option or not.
    """

    def __init__(
        self,
        name: str | None = None,
        description: str | None = None,
        name_localizations: dict[str, str] | None = None,
        description_localizations: dict[str, str] | None = None,
        required: bool | None = None,
        choices: list[Choice] | None = None,
        channel_types: list[ChannelType] = list(),
        min_value: int | None = None,
        max_value: int | None = None,
        autocomplete: bool | None = None,
    ) -> None:
        self.name = name
        self.description = description or "..."
        self.name_localizations = name_localizations
        self.description_localizations = description_localizations
        self.required = required
        self.choices = choices
        self.channel_types = channel_types
        self.min_value = min_value
        self.max_value = max_value
        self.autocomplete = autocomplete
        self.type: OptionType = OptionType.STRING

    def __repr__(self) -> str:
        return f"<Option name={self.name!r}>"

    def to_dict(self) -> dict[str, Any]:
        """
        Creates a dictionary representation of the option.

        Returns
        -------
        dict[str, Any]
            The dictionary representation of the option.
        """
        base: dict[str, Any] = {
            "name": self.name,
            "description": self.description,
            "type": self.type.value,
        }
        if self.name_localizations is not None:
            base["name_localizations"] = self.name_localizations
        if self.description_localizations is not None:
            base["description_localizations"] = self.description_localizations
        if self.required is not None:
            base["required"] = self.required
        if self.choices is not None:
            base["choices"] = [choice.to_dict() for choice in self.choices]
        if self.channel_types:
            base["channel_types"] = [t.value for t in self.channel_types]
        if self.min_value is not None:
            base["min_value"] = self.min_value
        if self.max_value is not None:
            base["max_value"] = self.max_value
        if self.autocomplete is not None:
            base["autocomplete"] = self.autocomplete
        return base


class Choice:
    """
    Represents a slash command choice.

    Attributes
    ----------
    name (str):
        The name of the choice.
    value (str):
        The value of the choice.
    name_localizations (dict[str, Any]|None):
        A dictionary of localized names for the choice, keyed by language code.
    """

    def __init__(
        self,
        name: str,
        name_localizations: dict[Locale, str] | None = None,
        value: str | int | float | bool | None = None,
    ) -> None:
        self.name = name
        self.value = value or name
        self.name_localizations = name_localizations

    def to_dict(self) -> dict[str, Any]:
        base: dict[str, Any] = {"name": self.name, "value": self.value}
        if self.name_localizations is not None:
            base["name_localizations"] = _populate_locales(self.name_localizations)
        return base



from __future__ import annotations

from enum import IntFlag
from typing import Any

__all__ = ("UserFlags", "GuildMemberFlags", "BaseFlag")


class BaseFlag(IntFlag):
    @classmethod
    def _missing_(cls, value: object) -> Any:
        if isinstance(value, str):
            if value.isdigit():
                return super()._missing_(int(value))


class UserFlags(IntFlag):
    STAFF = 1 << 0
    PARTNER = 1 << 1
    HYPESQUAD = 1 << 2
    BUG_HUNTER_LEVEL_1 = 1 << 3
    HYPESQUAD_ONLINE_HOUSE_1 = 1 << 6
    HYPESQUAD_ONLINE_HOUSE_2 = 1 << 7
    HYPESQUAD_ONLINE_HOUSE_3 = 1 << 8
    PREMIUM_EARLY_SUPPORTER = 1 << 9
    TEAM_PSEUDO_USER = 1 << 10
    BUG_HUNTER_LEVEL_2 = 1 << 14
    VERIFIED_BOT = 1 << 16
    VERIFIED_DEVELOPER = 1 << 17
    CERTIFIED_MODERATOR = 1 << 18
    BOT_HTTP_INTERACTIONS = 1 << 19
    ACTIVE_DEVELOPER = 1 << 22


class GuildMemberFlags(IntFlag):
    DID_REJOIN = 1 << 0
    COMPLETED_ONBOARDING = 1 << 1
    BYPASSES_VERIFICATION = 1 << 2
    STARTED_ONBOARDING = 1 << 3



from __future__ import annotations
import uuid
from typing import Any, Dict, Optional, TYPE_CHECKING

from dismake.types import AsyncFunction
from ..enums import ComponentType

if TYPE_CHECKING:
    from .view import View
    from ..models import Interaction

__all__ = ("Component",)


class Component:
    """
    Represents a Discord component.

    This class should not be initialized directly.
    All other components should subclass this class.

    Parameters
    ----------
    type: :class:`ComponentType`
        The type of the component.
    custom_id: :class:`str`
        The custom ID of the component. If not provided, a random UUID will be generated.
    disabled: :str:`bool`
        Indicates whether the component is disabled.

    Attributes
    ----------
    view: :class:`View`
        The view associated with the component.
    """

    def __init__(
        self, type: ComponentType, custom_id: Optional[str], disabled: Optional[bool]
    ) -> None:
        self.type = type
        self.custom_id = custom_id or str(uuid.uuid4())
        self.disabled = disabled
        self._view: View
        self._callback: AsyncFunction | None = None

    @property
    def view(self) -> View:
        """
        Returns the view associated with the component.
        """
        return self._view

    @view.setter
    def view(self, v: View) -> View:
        self.view = v
        return self.view

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts a component into a dict.

        Returns
        -------
        dict[str, Any]
        """
        base = {"type": self.type.value, "custom_id": self.custom_id}
        if self.disabled:
            base["disabled"] = self.disabled
        return base



from __future__ import annotations

from typing import Optional, List, Any, Dict
from datetime import datetime
from pydantic import BaseModel, validator
from .user import Member, User
from ..types import SnowFlake
from enum import Enum
from ..asset import Asset


class Ban(BaseModel):
    reason: str
    user: User


class GuildWidget(BaseModel):
    enabled: bool
    channel_id: SnowFlake


class GuildWidgetImageStyle(Enum):
    shield = "shield"
    banner1 = "banner1"
    banner2 = "banner2"
    banner3 = "banner3"
    banner4 = "banner4"


class WelcomeChannel(BaseModel):
    channel_id: SnowFlake
    description: str
    emoji_id: Optional[SnowFlake]
    emoji_name: str


class WelcomeScreen(BaseModel):
    description: str
    welcome_channels: List[WelcomeChannel]


class Guild(BaseModel):
    id: SnowFlake
    name: str
    icon: Optional[str]
    afk_channel_id: Optional[SnowFlake]
    afk_timeout: Optional[int]
    application_command_count: Optional[int]
    application_command_counts: Optional[Dict[str, int]]
    application_id: Optional[SnowFlake]
    banner: Optional[str]
    # channels: Dict[SnowFlake, Channel] = dict() TODO: Channel
    # default_message_notifications: GuildMessageNotification TODO
    description: Optional[str]
    discovery_splash: Optional[str]
    embedded_activities: List[Any] = []
    # emojis: Dict[SnowFlake, Emoji] TODO: Emoji
    # explicit_content_filter: ExplicitContentFilterLevel TODO
    features: List[str]
    guild_hashes: Dict[Any, Any] = {}
    # guild_scheduled_events: Dict[SnowFlake, GuildScheduledEvent] = {} TODO GuildSheduledEvent
    hub_type: Optional[bool]
    joined_at: Optional[datetime]
    large: bool = False
    lazy: Optional[bool]
    max_members: int
    max_video_channel_users: Optional[int]
    member_count: int = 0
    members: Dict[SnowFlake, Member] = {}
    # mfa_level: MFALevel TODO
    nsfw: bool
    # nsfw_level: NSFWLevel TODO
    owner_id: SnowFlake
    preferred_locale: str
    premium_progress_bar_enabled: Optional[bool]
    premium_subscription_count: int
    # premium_tier: PremiumTierLevel TODO
    presences: Optional[List[Dict[str, Any]]]
    public_updates_channel_id: Optional[SnowFlake]
    # roles: Dict[SnowFlake, Role] TODO: Role
    rules_channel_id: Optional[SnowFlake]
    splash: Optional[str]
    stage_instances: Optional[List[Any]] = list()
    # stickers: Optional[Dict[SnowFlake, Sticker]] = dict() TODO Sticker
    system_channel_flags: Optional[int]
    system_channel_id: Optional[SnowFlake]
    # threads: Optional[Dict[SnowFlake, Thread]] = dict() TODO: Thread
    unavailable: Optional[bool]
    vanity_url_code: Optional[str]
    # verification_level: VerificationLevel TODO
    voice_states: Optional[List[Any]] = list()
    created_at: Optional[datetime]

    @property
    def display_banner(self) -> Optional[Asset]:
        if self.banner:
            return Asset.from_guild_banner(self.banner, int(self.id))
        return None

    @property
    def display_icon(self) -> Optional[Asset]:
        if self.icon:
            return Asset.from_guild_icon(self.icon, int(self.id))
        return None



from __future__ import annotations
from typing import Any, Dict, List, Optional, Union, TYPE_CHECKING

from .enums import MessageFlags

if TYPE_CHECKING:
    from .ui import View


__all__ = ("handle_send_params", "handle_edit_params")


def handle_send_params(
    content: str,
    *,
    tts: Optional[bool] = None,
    embeds: Optional[list] = None,
    allowed_mentions: Optional[Any] = None,
    view: Optional[Union[View, Dict[Any, Any]]] = None,
    attachments: Optional[List[Any]] = None,
    embed: Optional[Any] = None,
    ephemeral: bool = False,
):
    payload: dict[str, Any] = {"content": content}
    if ephemeral:
        payload.update({"flags": MessageFlags.EPHEMERAL.value})
    if tts:
        payload.update({"tts": tts})
    if view:
        if isinstance(view, dict):
            payload.update({"components": view})
        else:
            payload.update({"components": view.to_dict()})
    _embeds = list()
    if embeds:
        for emb in embeds:
            _embeds.append(emb)
    if embed:
        _embeds.append(embed)

    if _embeds:
        embeds = _embeds
    return payload


def handle_edit_params(
    content: str,
    *,
    tts: Optional[bool] = None,
    embeds: Optional[list] = None,
    allowed_mentions: Optional[Any] = None,
    view: Optional[Union[View, Dict[str, Any]]] = None,
    attachments: Optional[List[Any]] = None,
    embed: Optional[Any] = None,
) -> Dict[str, Any]:
    payload: Dict[str, Any] = {"content": content}
    if tts:
        payload.update({"tts": tts})
    embs = list()
    if embeds:
        for embed in embeds:
            embs.append(embed)
    if embed:
        embs.append(embed)

    if embeds:
        embeds = embs

    if view:
        if isinstance(view, dict):
            payload.update({"components": view})
        else:
            payload.update({"components": view.to_dict()})
    else:
        payload.update({"components": None})
    return payload



from __future__ import annotations
from re import I

from typing import Any, List, Optional, TYPE_CHECKING, Union, Dict, TYPE_CHECKING
from pydantic import BaseModel
from .user import Member, User
from .guild import Guild
from .role import Role
from .message import Message
from ..enums import InteractionType, InteractionResponseType, MessageFlags, OptionType
from ..errors import InteractionNotResponded, InteractionResponded
from ..params import handle_send_params, handle_edit_params
from fastapi import Request
from ..types import SnowFlake
from .channels import Channel
from .components import TextInput

if TYPE_CHECKING:
    from ..ui import View, Modal
    from ..client import Bot
    from ..commands import Choice


__all__ = (
    "Interaction",
    "ApplicationCommandData",
    "ApplicationCommandOption",
    "MessageComponentData",
    "ModalSubmitData",
    "ModalSubmitActionRowData",
)


def _extract_options(
    option: ApplicationCommandOption,
) -> List[ApplicationCommandOption]:
    """Recursively extract options from commands"""
    if option.type == OptionType.SUB_COMMAND.value and option.options is not None:
        return [o for sub_opt in option.options for o in _extract_options(sub_opt)]
    elif (
        option.type == OptionType.SUB_COMMAND_GROUP.value and option.options is not None
    ):
        return [o for sub_group in option.options for o in _extract_options(sub_group)]
    else:
        return [option]


def _options_to_dict(
    options: List[ApplicationCommandOption], resolved_data: Optional[ResolvedData]
) -> Dict[str, Any]:
    """Convert options to a dictionary"""
    namespace_dict = {}
    for option in options:
        if option.type == OptionType.USER.value and resolved_data is not None:
            if resolved_data.users is not None:
                namespace_dict[option.name.replace("-", "_")] = resolved_data.users.get(
                    str(option.value)  # type: ignore
                )
        elif option.type == OptionType.ROLE.value and resolved_data is not None:
            if resolved_data.roles is not None:
                namespace_dict[option.name.replace("-", "_")] = resolved_data.roles.get(
                    str(option.value)  # type: ignore
                )
        elif option.type == OptionType.CHANNEL.value and resolved_data is not None:
            if resolved_data.channels is not None:
                namespace_dict[
                    option.name.replace("-", "_")
                ] = resolved_data.channels.get(
                    str(option.value)
                )  # type: ignore
        else:
            namespace_dict[option.name.replace("-", "_")] = option.value  # type: ignore
    return namespace_dict


class ResolvedData(BaseModel):
    users: Optional[Dict[str, User]]
    members: Optional[Any]
    channels: Optional[Dict[str, Channel]]
    roles: Optional[Dict[str, Role]]
    messages: Optional[Any]
    attachments: Optional[Any]


class ApplicationCommandOption(BaseModel):
    name: str
    type: int
    value: Optional[Union[str, int, float, bool, User, Role, Channel]]
    options: Optional[List[ApplicationCommandOption]]
    focused: bool = False


class ApplicationCommandData(BaseModel):
    id: SnowFlake
    name: str
    type: int
    resolved: Optional[ResolvedData]
    options: Optional[List[ApplicationCommandOption]]
    guild_id: Optional[SnowFlake]
    target_id: Optional[SnowFlake]


class MessageComponentData(BaseModel):
    custom_id: str
    component_type: int
    values: Optional[List[str]]


class ModalSubmitActionRowData(BaseModel):
    components: list[TextInput]


class ModalSubmitData(BaseModel):
    custom_id: str
    components: list[ModalSubmitActionRowData]


class Interaction:
    """
    Represents a Discord interaction.

    An interaction happens when a user does an action that needs to
    be notified. Current examples are slash commands and components.

    Parameters
    -----------
    request: Request
        The request object.
    data: Dict[str, Any]
        The interaction data
    """

    __slots__ = (
        "_request",
        "_is_response_done",
        "id",
        "application_id",
        "type",
        "token",
        "version",
        "guild_id",
        "channel_id",
        "app_permissions",
        "locale",
        "guild_locale",
        "user",
        "_data",
        "data",
        "channel",
        "__message",
        "message",
    )

    def __init__(self, request: Request, data: Dict[str, Any]) -> None:
        self._request = request
        self._is_response_done = False
        self.id: int = int(data["id"])
        self.application_id: SnowFlake = data["application_id"]
        self.type: int = data["type"]
        self.token: str = data["token"]
        self.version: int = data["version"]
        self.guild_id: Optional[int] = data.get("guild_id")
        self.channel_id: Optional[SnowFlake] = data.get("channel_id")
        self.app_permissions: Optional[int] = data.get("app_permissions")
        self.locale: Optional[str] = data.get("locale")
        self.guild_locale: Optional[str] = data.get("guild_locale")
        self.user: Union[User, Member]
        self._data: Optional[dict] = data.get("data")
        self.data: Optional[
            Union[ApplicationCommandData, ModalSubmitData, MessageComponentData]
        ]
        if self.is_application_command or self.is_autocomplete:
            self.data = ApplicationCommandData(**self._data) if self._data else None
        elif self.is_message_component:
            self.data = MessageComponentData(**self._data) if self._data else None
        else:
            self.data = ModalSubmitData(**self._data) if self._data else None
        if self.guild_id:
            try:
                member = data["member"]
            except:
                pass
            else:
                self.user = Member(**member)
        else:
            self.user = User(**data["user"])
        self.channel: Optional[Any] = data.get("channel")
        self.__message: Optional[dict] = data.get("message")
        self.message: Optional[Message]
        if self.__message is not None:
            self.message = Message(_request=self._request, **self.__message)

    @property
    def bot(self) -> Bot:
        """
        :class:`Bot`: The bot that is handling this interaction.
        """
        return self._request.app

    @property
    def is_application_command(self) -> bool:
        """
        The is_application_command function checks if the interaction type is an application command.
        """
        return self.type == InteractionType.APPLICATION_COMMAND.value

    @property
    def is_autocomplete(self) -> bool:
        """
        The is_autocomplete function checks if the interaction type is an autocomplete.
        """
        return self.type == InteractionType.APPLICATION_COMMAND_AUTOCOMPLETE.value

    @property
    def is_modal_submit(self) -> bool:
        """
        The is_modal_submit function checks if the interaction type is a modal submit.
        """
        return self.type == InteractionType.MODAL_SUBMIT.value

    @property
    def is_ping(self) -> bool:
        """
        The is_ping function checks if the interaction type is a ping.
        """
        return self.type == InteractionType.PING.value

    @property
    def is_message_component(self) -> bool:
        """
        The is_component function checks if the interaction is a component.
        """
        return self.type == InteractionType.MESSAGE_COMPONENT.value

    @property
    def is_responded(self) -> bool:
        """
        :class:`bool`: Indicates whether an interaction response has been done before.
        An interaction can only be responded to once.
        """
        return self._is_response_done

    @property
    def namespace(self) -> Namespace:
        if not isinstance(self.data, ApplicationCommandData):
            return Namespace(**{})
        if (data := self.data) is None or (options := data.options) is None:
            return Namespace(**{})

        options = [o for option in options for o in _extract_options(option)]
        if not options:
            return Namespace(**{})

        return Namespace(**_options_to_dict(options, data.resolved))

    async def fetch_guild(self) -> Optional[Guild]:
        """
        This function is a coroutine that returns the Guild object for this interaction.
        It is used to get the guild object from its ID, which is stored in self.guild_id.

        Returns
        -------
        Optional[Guild]:
            A guild object if the guild_id is not none.
        """
        if self.guild_id is None:
            return None
        return await self.bot.fetch_guild(self.guild_id)

    async def respond(
        self,
        content: str,
        *,
        tts: bool = False,
        ephemeral: bool = False,
        view: Optional[View] = None,
    ):
        if self.is_responded:
            raise InteractionResponded(self)

        if view:
            self.bot.add_view(view)
        await self.bot._http.client.request(
            method="POST",
            url=f"/interactions/{self.id}/{self.token}/callback",
            json={
                "type": InteractionResponseType.CHANNEL_MESSAGE_WITH_SOURCE.value,
                "data": handle_send_params(
                    content=content, tts=tts, ephemeral=ephemeral, view=view
                ),
            },
            headers=self.bot._http.headers,
        )
        self._is_response_done = True

    async def defer(self, thinking: bool = True):
        if self.is_responded:
            raise InteractionResponded(self)
        await self.bot._http.client.request(
            method="POST",
            url=f"/interactions/{self.id}/{self.token}/callback",
            json={
                "type": InteractionResponseType.DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE.value,
                "data": {"flags": MessageFlags.LOADING.value} if not thinking else None,
            },
            headers=self.bot._http.headers,
        )
        self._is_response_done = True

    async def send_followup(
        self,
        content: str,
        *,
        tts: bool = False,
        view: Optional[View] = None,
        ephemeral: bool = False,
    ):
        if not self.is_responded:
            raise InteractionNotResponded(self)

        if view:
            self.bot.add_view(view)
        return await self.bot._http.client.request(
            method="POST",
            url=f"/webhooks/{self.application_id}/{self.token}",
            json=handle_send_params(
                content=content, tts=tts, view=view, ephemeral=ephemeral
            ),
        )

    async def edit_original_response(
        self, content: str, *, tts: bool = False, view: Optional[View] = None
    ):
        if view:
            self.bot.add_view(view)
        return await self.bot._http.client.request(
            method="PATCH",
            url=f"/webhooks/{self.application_id}/{self.token}/messages/@original",
            json=handle_edit_params(content=content, tts=tts, view=view),
        )

    async def get_original_response(self) -> Message:
        res = await self.bot._http.client.request(
            method="GET",
            url=f"/webhooks/{self.application_id}/{self.token}/messages/@original",
        )
        res.raise_for_status()
        return Message(**res.json())

    async def send(
        self,
        content: str,
        *,
        tts: bool = False,
        view: Optional[View] = None,
        ephemeral: bool = False,
    ):
        if self.is_responded:
            return await self.send_followup(
                content, tts=tts, view=view, ephemeral=ephemeral
            )
        return await self.respond(content, tts=tts, view=view, ephemeral=ephemeral)

    async def edit_message(
        self, content: str, *, tts: bool = False, view: Optional[View] = None
    ):
        if not self.is_message_component:
            return
        if self.is_responded:
            raise InteractionResponded(self)
        if view:
            self.bot.add_view(view)
        payload = handle_edit_params(content=content, tts=tts, view=view)
        return await self.bot._http.client.request(
            method="POST",
            url=f"/interactions/{self.id}/{self.token}/callback",
            json={
                "type": InteractionResponseType.UPDATE_MESSAGE.value,
                "data": payload,
            },
        )

    async def autocomplete(self, choices: List[Choice]):
        if not self.is_autocomplete:
            return

        return await self.bot._http.client.request(
            method="POST",
            url=f"/interactions/{self.id}/{self.token}/callback",
            json={
                "type": InteractionResponseType.APPLICATION_COMMAND_AUTOCOMPLETE_RESULT.value,
                "data": {"choices": [choice.to_dict() for choice in choices]},
            },
        )

    async def respond_with_modal(self, modal: Modal):
        if self.is_responded:
            raise InteractionResponded(self)
        self.bot.add_modal(modal)
        return await self.bot._http.client.request(
            method="POST",
            url=f"/interactions/{self.id}/{self.token}/callback",
            json={"type": InteractionResponseType.MODAL.value, "data": modal.to_dict()},
        )


class Namespace:
    """
    Inspired from discord.py
    """

    def __init__(self, **kwargs) -> None:
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __getattr__(self, value) -> None:
        return None



from __future__ import annotations

from typing import Optional, Any, Dict, Union
from .component import Component
from ..enums import ButtonStyles, ComponentType
from ..models import PartialEmoji

__all__ = ("Button",)


class Button(Component):
    """
    Represents a discord button.

    Parameters
    ----------
    label: :class:`str`
        The text displayed on the button.
    custom_id: :class:`str`
        A unique identifier for the button.
    style: :class:`ButtonStyles`
        The style of the button
    emoji: Union[:class:`Emoji`, :class:`str`]
        The emoji displayed on the button.
    url: :class:`str`
        If set, this button will act as a link to the given URL.
    disabled: :class:`bool`
        If True, the button will be disabled and cannot be clicked.

    Attributes
    ----------
    view: :class:`View`
        The view associated with the component.
    """

    def __init__(
        self,
        label: Optional[str],
        custom_id: Optional[str],
        style: Optional[ButtonStyles],
        emoji: Optional[Union[PartialEmoji, str]] = None,
        url: Optional[str] = None,
        disabled: Optional[bool] = None,
    ) -> None:
        self.type = ComponentType.BUTTON
        super().__init__(type=self.type, custom_id=custom_id, disabled=disabled)
        self.label = label
        self.style = style or ButtonStyles.PRIMARY
        self.url = url
        self.emoji: Optional[PartialEmoji]
        if emoji:
            if isinstance(emoji, PartialEmoji):
                self.emoji = emoji
            else:
                self.emoji = PartialEmoji.from_str(emoji)
        else:
            self.emoji = None

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the button into dict.

        Returns
        -------
        dict[str, Any]
        """
        base = super().to_dict()
        if not self.url:
            base["label"] = self.label
            base["style"] = self.style.value
            if self.emoji is not None:
                base["emoji"] = self.emoji.dict(exclude_none=True)
        else:
            base["label"] = self.label
            base["style"] = ButtonStyles.LINK.value
            base["url"] = self.url
            if self.emoji is not None:
                base["emoji"] = self.emoji.dict(exclude_none=True)
            if base.get("custom_id"):
                base.pop("custom_id")

        return base



from __future__ import annotations
from typing import Any, Optional, Union

from pydantic import BaseModel, validator
from ..enums import ComponentType, TextInputStyle
from .emoji import PartialEmoji

__all__ = ("SelectOption", "Component", "ActionRow", "TextInput")


class Component(BaseModel):
    type: ComponentType
    custom_id: Optional[str] = None


class ActionRow(Component):
    type: ComponentType = ComponentType.ACTION_ROW
    components: list[Component] = list()

    @validator("components")
    def _validate_components(cls, value: list[Component]) -> list[Component]:
        if len(value) > 5:
            raise ValueError("An action row can only have 5 components.")
        if ComponentType.ACTION_ROW is [t.type for t in value]:
            raise ValueError("A action row can not have another action row.")

        if (
            ComponentType.STRING_SELECT
            or ComponentType.ROLE_SELECT
            or ComponentType.CHANNEL_SELECT
            or ComponentType.MENTIONABLE_SELECT
            and ComponentType.BUTTON in ([t.type for t in value])
        ):
            raise ValueError(
                "An action row can not contains select option and button option option together."
            )

        return value

    @property
    def is_full(self) -> bool:
        return len(self.components) == 5


class SelectOption(BaseModel):
    label: str
    value: str
    description: Optional[str]
    emoji: Optional[Union[PartialEmoji, dict, str]]
    default: Optional[bool]


class TextInput(Component):
    value: Optional[str]



from __future__ import annotations
from typing import Literal
from pydantic import BaseModel
from ..permissions import Permissions
from ..types import SnowFlake

__all__ = ("PermissionOverwrites",)


class PermissionOverwrites(BaseModel):
    id: SnowFlake
    type: Literal[0, 1]
    allow: Permissions
    deny: Permissions



from __future__ import annotations


__all__ = ("Asset",)


class AssetMixin:
    def __init__(self):
        pass


class Asset(AssetMixin):
    """Represent a CDN asset on Discord."""

    BASE = "https://cdn.discordapp.com"

    def __init__(self, *, url: str, key: str, animated: bool = False) -> None:
        self._url = url
        self._animated = animated
        self._key = key

    def __str__(self) -> str:
        return self._url

    @classmethod
    def from_default_avatar(cls, index: int) -> Asset:
        return cls(
            url=f"{cls.BASE}/embed/avatars/{index}.png", key=str(index), animated=False
        )

    @classmethod
    def from_avatar(cls, avatar_hash: str, user_id: int):
        animated = avatar_hash.startswith("a_")
        format = "gif" if animated else "png"
        return cls(
            url=f"{cls.BASE}/avatars/{user_id}/{avatar_hash}.{format}/?size=1024",
            animated=animated,
            key=avatar_hash,
        )

    @classmethod
    def from_guild_banner(cls, banner_hash: str, guild_id: int):
        animated = banner_hash.startswith("a_")
        format = "gif" if animated else "png"
        return cls(
            url=f"{cls.BASE}/banners/{guild_id}/{banner_hash}.{format}",
            animated=animated,
            key=banner_hash,
        )

    @classmethod
    def from_guild_icon(cls, icon_hash: str, guild_id: int):
        animated = icon_hash.startswith("a_")
        format = "gif" if animated else "png"
        return cls(
            url=f"{cls.BASE}/icons/{guild_id}/{icon_hash}.{format}",
            key=icon_hash,
            animated=animated,
        )



from __future__ import annotations

from enum import Enum, IntEnum

__all__ = (
    "DefaultAvatar",
    "InteractionType",
    "InteractionResponseType",
    "InteractionResponseFlags",
    "CommandType",
    "OptionType",
    "MessageFlags",
    "ChannelType",
    "StrEnum",
    "Locale",
    "ComponentType",
    "ButtonStyles",
    "TextInputStyle",
)


class StrEnum(str, Enum):
    def __str__(self) -> str:
        return self.value


class DefaultAvatar(Enum):
    blurple = 0
    grey = 1
    gray = 1
    green = 2
    orange = 3
    red = 4

    def __str__(self) -> str:
        return self.name


class InteractionType(Enum):
    PING = 1
    APPLICATION_COMMAND = 2
    MESSAGE_COMPONENT = 3
    APPLICATION_COMMAND_AUTOCOMPLETE = 4
    MODAL_SUBMIT = 5


class InteractionResponseType(Enum):
    PONG = 1
    CHANNEL_MESSAGE_WITH_SOURCE = 4
    DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE = 5
    DEFERRED_UPDATE_MESSAGE = 6
    UPDATE_MESSAGE = 7
    APPLICATION_COMMAND_AUTOCOMPLETE_RESULT = 8
    MODAL = 9


class InteractionResponseFlags(Enum):
    EPHEMERAL = 1 << 6


class CommandType(Enum):
    SLASH = 1
    USER = 2
    MESSAGE = 3


class OptionType(Enum):
    SUB_COMMAND = 1
    SUB_COMMAND_GROUP = 2
    STRING = 3
    INTEGER = 4
    BOOLEAN = 5
    USER = 6
    CHANNEL = 7
    ROLE = 8
    MENTIONABLE = 9
    NUMBER = 10


class MessageFlags(Enum):
    CROSSPOSTED = 1 << 0
    IS_CROSSPOST = 1 << 1
    SUPPRESS_EMBEDS = 1 << 2
    SOURCE_MESSAGE_DELETED = 1 << 3
    URGENT = 1 << 4
    HAS_THREAD = 1 << 5
    EPHEMERAL = 1 << 6
    LOADING = 1 << 7
    FAILED_TO_MENTION_SOME_ROLES_IN_THREAD = 1 << 8
    SUPPRESS_NOTIFICATIONS = 1 << 12


class ChannelType(Enum):
    GUILD_TEXT = 0
    DM = 1
    GUILD_VOICE = 2
    GROUP_DM = 3
    GUILD_CATEGORY = 4
    GUILD_ANNOUNCEMENT = 5
    ANNOUNCEMENT_THREAD = 10
    PUBLIC_THREAD = 11
    PRIVATE_THREAD = 12
    GUILD_STAGE_VOICE = 13
    GUILD_DIRECTORY = 14
    GUILD_FORUM = 15


class Locale(StrEnum):
    id = "id"  # Indonesian   Bahasa Indonesia
    da = "da"  # Danish	Dansk
    de = "de"  # 	German	Deutsch
    en_GB = "en-GB"  # 	English, UK	English, UK
    en_US = "en-US"  # 	English, US	English, US
    es_ES = "es-ES"  # 	Spanish	Español
    fr = "fr"  # 	French	Français
    hr = "hr"  # 	Croatian	Hrvatski
    it = "it"  # 	Italian	Italiano
    lt = "lt"  # 	Lithuanian	Lietuviškai
    hu = "hu"  # 	Hungarian	Magyar
    nl = "nl"  # 	Dutch	Nederlands
    no = "no"  # 	Norwegian	Norsk
    pl = "pl"  # 	Polish	Polski
    pt_BR = "pt-BR"  # 	Portuguese, Brazilian	Português do Brasil
    ro = "ro"  # 	Romanian, Romania	Română
    fi = "fi"  # 	Finnish	Suomi
    sv_SE = "sv-SE"  # 	Swedish	Svenska
    vi = "vi"  # 	Vietnamese	Tiếng Việt
    tr = "tr"  # 	Turkish	Türkçe
    cs = "cs"  # 	Czech	Čeština
    el = "el"  # 	Greek	Ελληνικά
    bg = "bg"  # 	Bulgarian	български
    ru = "ru"  # 	Russian	Pусский
    uk = "uk"  # 	Ukrainian	Українська
    hi = "hi"  # 	Hindi	हिन्दी
    th = "th"  # 	Thai	ไทย
    zh_CN = "zh-CN"  # 	Chinese, China	中文
    ja = "ja"  # 	Japanese	日本語
    zh_TW = "zh-TW"  # 	Chinese, Taiwan	繁體中文
    ko = "ko"  # 	Korean	한국어


class ComponentType(Enum):
    ACTION_ROW = 1
    BUTTON = 2
    STRING_SELECT = 3
    TEXT_INPUT = 4
    USER_SELECT = 5
    ROLE_SELECT = 6
    MENTIONABLE_SELECT = 7
    CHANNEL_SELECT = 8


class ButtonStyles(Enum):
    PRIMARY = primary = 1
    SECONDARY = secondary = 2
    SUCCESS = success = 3
    DANGER = danger = 4
    LINK = link = 5


class TextInputStyle(Enum):
    SHORT = short = 1
    PARAGRAPH = paragraph = 2



from __future__ import annotations
import logging
from typing import Iterable

__all__ = ("chunk",)

format = "%(asctime)s %(name)-15s | %(message)s"

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": True,
    "handlers": {
        "console": {
            "class": "rich.logging.RichHandler",
            "formatter": "default",
            "show_time": False,
            "rich_tracebacks": True,
        },
    },
    "formatters": {
        "default": {
            "format": format,
        },
    },
    "loggers": {
        "uvicorn": {
            "handlers": ["console"],
            "level": "DEBUG",
            "propagate": True,
        },
    },
}


def chunk(max_size: int, iterator: Iterable):
    if max_size <= 0:
        raise ValueError("Chunk sizes must be greater than 0.")

    ret = list()
    n = 0
    for i in iterator:
        ret.append(i)
        n += 1
        if n == max_size:
            yield ret
            ret.clear()
            n = 0
    if ret:
        yield ret



from __future__ import annotations

import re
from typing import Optional
from pydantic import BaseModel
from ..types import SnowFlake
from .user import User


__all__ = ("Emoji", "PartialEmoji")


class PartialEmoji(BaseModel):
    id: Optional[SnowFlake]
    name: Optional[str]
    animated: Optional[bool]

    @classmethod
    def from_str(cls, value: str):
        CUSTOM_EMOJI_RE = re.compile(
            r"<?(?P<animated>a)?:?(?P<name>[A-Za-z0-9\_]+):(?P<id>[0-9]{13,20})>?"
        )
        match = CUSTOM_EMOJI_RE.match(value)
        if match is not None:
            groups = match.groupdict()
            animated = bool(groups["animated"])
            id = int(groups["id"])
            name = groups["name"]
            return cls(id=id, name=name, animated=animated)
        return cls(id=None, name=value, animated=None)


class Emoji(BaseModel):
    roles: Optional[list[int]]
    user: Optional[User]
    require_colons: Optional[bool]
    managed: Optional[bool]
    available: Optional[bool]



from setuptools import setup, find_packages

with open("README.md", "r") as file:
    long_description = file.read()


setup(
    name="dismake",
    version="0.0.17",
    author="Pranoy Majumdar",
    author_email="officialpranoy2@gmail.com",
    description="None",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Apptic-Development/dismake",
    project_urls={"Homepage": "https://github.com/Apptic-Development/dismake"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    python_requires=">=3.9",
    license="MIT",
    entry_points={"console_scripts": ["dismake=dismake.cli:main"]},
    package_data={"templates": ["templates/*"]},
    install_requires=[
        "fastapi==0.95.1",
        "httpx==0.24.0",
        "pydantic==1.10.7",
        "PyNaCl==1.5.0",
        "rich==13.3.5",
        "typing_extensions==4.5.0",
        "uvicorn==0.22.0",
    ],
)


print("Successfully installed dismake")



from __future__ import annotations

from .flags import BaseFlag

__all__ = ("Permissions",)


class Permissions(BaseFlag):
    CREATE_INSTANT_INVITE = 1 << 0  # Allows creation of instant invites (T, V, S)
    KICK_MEMBERS = 1 << 1  # Allows kicking members
    BAN_MEMBERS = 1 << 2  # Allows banning members
    ADMINISTRATOR = (
        1 << 3
    )  # Allows all permissions and bypasses channel permission overwrites
    MANAGE_CHANNELS = 1 << 4  # Allows management and editing of channels (T, V, S)
    MANAGE_GUILD = 1 << 5  # Allows management and editing of the guild
    ADD_REACTIONS = 1 << 6  # Allows for the addition of reactions to messages (T, V, S)
    VIEW_AUDIT_LOG = 1 << 7  # Allows for viewing of audit logs
    PRIORITY_SPEAKER = (
        1 << 8
    )  # Allows for using priority speaker in a voice channel (V)
    STREAM = 1 << 9  # Allows the user to go live (V, S)
    VIEW_CHANNEL = (
        1 << 10
    )  # Allows guild members to view a channel, which includes reading messages in text channels and joining voice channels (T, V, S)
    SEND_MESSAGES = (
        1 << 11
    )  # Allows for sending messages in a channel and creating threads in a forum (does not allow sending messages in threads) (T, V, S)
    SEND_TTS_MESSAGES = 1 << 12  # Allows for sending of /tts messages (T, V, S)
    MANAGE_MESSAGES = 1 << 13  # Allows for deletion of other users' messages (T, V, S)
    EMBED_LINKS = (
        1 << 14
    )  # Links sent by users with this permission will be auto-embedded (T, V, S)
    ATTACH_FILES = 1 << 15  # Allows for uploading images and files (T, V, S)
    READ_MESSAGE_HISTORY = 1 << 16  # Allows for reading of message history (T, V, S)
    MENTION_EVERYONE = (
        1 << 17
    )  # Allows for using the @everyone tag to notify all users in a channel, and the @here tag to notify all online users in a channel (T, V, S)
    USE_EXTERNAL_EMOJIS = (
        1 << 18
    )  # Allows the usage of custom emojis from other servers (T, V, S)
    VIEW_GUILD_INSIGHTS = 1 << 19  # Allows for viewing guild insights
    CONNECT = 1 << 20  # Allows for joining of a voice channel (V, S)
    SPEAK = 1 << 21  # Allows for speaking in a voice channel (V)
    MUTE_MEMBERS = 1 << 22  # Allows for muting members in a voice channel (V, S)
    DEAFEN_MEMBERS = 1 << 23  # Allows for deafening of members in a voice channel (V)
    MOVE_MEMBERS = 1 << 24  # Allows for moving of members between voice channels (V, S)
    USE_VAD = (
        1 << 25
    )  # Allows for using voice-activity-detection in a voice channel (V)
    CHANGE_NICKNAME = 1 << 26  # Allows for modification of own nickname
    MANAGE_NICKNAMES = 1 << 27  # Allows for modification of other users' nicknames



from __future__ import annotations

import argparse
import json
from pathlib import Path


# def generate_app(path: Path):
#     template = Path(pkg_resources.resource_filename("dismake", "templates"))
#     if not template.exists():
#         return print(
#             f"""
#     Error
# =============
# The templates folder is missing.
# You can fix this issue by reinstalling the package.

# Command:
# pip uninstall dismake -y && pip install dismake
# """
#         )
#     path.mkdir(parents=True, exist_ok=True)
#     for item in template.iterdir():
#         if item.is_file():
#             shutil.copyfile(item, path / item.name)
#         else:
#             shutil.copytree(item, path / item.name)
#     return path


def print_help(parser: argparse.ArgumentParser):
    parser.print_help()


# def init_command(args: argparse.Namespace):
#     project_name = input("Please enter a name for your project: ")
#     try:
#         path = Path(project_name)
#     except Exception as e:
#         return print(f"TODO")
#     else:
#         if path.is_file():
#             print(f"A file already exists with this name.")
#         created = generate_app(path, "basic")
#         if created:
#             return print(
#                 f"Successfully created your project in {created.absolute().name!r}"
#             )


def vercel_command(args):
    path = Path(".") / "vercel.json"
    config = {
        "builds": [{"src": "<your-main-file>", "use": "@vercel/python"}],
        "routes": [{"src": "/(.*)", "dest": "your-main-file"}],
    }
    path.touch()
    with open(path, "w") as f:
        f.write(json.dumps(config))
    print(
        f"""Successfully created a 'vercel.json' file.
Replace the '<your-main-file>' with your main file name example 'main.py'."""
    )


def add_subparsers(parser):
    subparsers = parser.add_subparsers()
    vercel = subparsers.add_parser(
        "vercel", help="Generate a vercel.json file for deployment on vercel."
    )
    vercel.set_defaults(func=vercel_command)

    # init = subparsers.add_parser("init", help="Creates a new dismake project for you.")
    # init.set_defaults(func=init_command)
    return subparsers


def main():
    parser = argparse.ArgumentParser(
        prog="dismake",
        description="The command line interface for dismake library.",
    )
    add_subparsers(parser)
    args = parser.parse_args()
    if not bool(args._get_kwargs()):
        return print_help(parser)
    return args.func(args)


if __name__ == "__main__":
    main()



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .commands import Command
    from .models import Interaction


__all__ = (
    "DismakeException",
    "NotImplemented",
    "CommandInvokeError",
    "InteractionResponded",
    "InteractionNotResponded",
    "ComponentException",
    "PluginException",
    "CommandException",
    "ModalException",
)


class DismakeException(Exception):
    """Base dismake exception."""


class PluginException(Exception):
    """Base plugin exception."""


class ModalException(Exception):
    """Base modal exception."""


class NotImplemented(DismakeException):
    """Raise when a unknown command invoke."""


class CommandInvokeError(DismakeException):
    """When a command fails to invoke."""

    def __init__(self, command: Command, exception: Exception) -> None:
        self.command = command
        self.exception = exception
        super().__init__(f"Command {command.name!r} raised an exception: {exception}")


class CommandException(DismakeException):
    """Raise when a command raised an exception."""


class InteractionResponded(DismakeException):
    def __init__(self, interaction: Interaction) -> None:
        self.interaction = interaction
        super().__init__(f"{interaction.id!r} Interaction already responded.")


class InteractionNotResponded(DismakeException):
    def __init__(self, interaction: Interaction) -> None:
        self.interaction = interaction
        super().__init__(f"{interaction.id!r} The interaction is not responded.")


class ComponentException(DismakeException):
    """Base Exception for View."""



from __future__ import annotations

from typing import Any, TYPE_CHECKING
from fastapi import Request, Response
from fastapi.responses import JSONResponse
from nacl.signing import VerifyKey
from nacl.exceptions import BadSignatureError
from .enums import InteractionType, InteractionResponseType
from .models import (
    Interaction,
    ApplicationCommandData,
    MessageComponentData,
    ModalSubmitData,
)
from .commands import Command, Group
from logging import getLogger

if TYPE_CHECKING:
    from .client import Bot

log = getLogger("uvicorn")


class InteractionHandler:
    """
    Handles interactions.

    Parameters
    ----------
    client: :class:`Bot`
        The bot object.
    """

    __slots__ = ("client", "verification_key")

    def __init__(self, client: Bot) -> None:
        self.client = client
        self.verification_key = VerifyKey(bytes.fromhex(client._client_public_key))

    def verify_key(self, body: bytes, signature: str, timestamp: str):
        """
        Verifies the signature of the request.

        Parameters
        ----------
        body: :class:`bytes`
            The body of the request.
        signature: :class:`str`
            The signature of the request.
        timestamp: :class:`str`
            The timestamp of the request.

        Returns
        -------
        Whether the signature is valid.
        """
        message = timestamp.encode() + body
        try:
            self.verification_key.verify(message, bytes.fromhex(signature))
            return True
        except BadSignatureError as e:
            log.error("Bad signature request.")
        except Exception as e:
            log.exception(e)
            return False

    async def _handle_command(self, request: Request) -> Any:
        """
        Handles a command request.

        Parameters
        ----------
        request: :class:`Request`
            The request object.

        Returns
        -------
        The response object.
        """
        payload: dict = await request.json()
        interaction = Interaction(request=request, data=payload)
        assert isinstance(interaction.data, ApplicationCommandData)
        if (data := interaction.data) is not None:
            command = self.client._commands.get(data.name)
            if command is not None:
                if isinstance(command, Command):
                    await command.invoke(interaction)
                elif isinstance(command, Group):
                    assert data.options is not None, "Invalid data recieved."
                    child_2 = command.commands.get(data.options[0].name)
                    if isinstance(child_2, Group):
                        assert (
                            data.options[0].options is not None
                        ), "Invalid data recieved."
                        child_3 = child_2.commands.get(data.options[0].options[0].name)
                        assert isinstance(
                            child_3, Command
                        ), f"command {child_3} is too nested."
                        await child_3.invoke(interaction)
                    if isinstance(child_2, Command):
                        await child_2.invoke(interaction)

    async def _handle_autocomplete(self, request: Request) -> Any:
        """
        Handles an autocomplete request.

        Parameters
        ----------
        request: :class:`Request`
            The request object.

        Returns
        -------
        The response object.
        """
        payload: dict = await request.json()
        interaction = Interaction(request=request, data=payload)

        if not (
            interaction.data is not None
            and isinstance(interaction.data, ApplicationCommandData)
            and interaction.is_autocomplete
        ):
            return
        if not (command := self.client.get_command(interaction.data.name)):
            return

        if isinstance(command, Command) and interaction.data.options is not None:
            options = interaction.data.options
            focused = list(filter(lambda x: x.focused == True, options))
            if not focused:
                raise ValueError("No focus items! Probably this is a discord bug.")
            return await command.invoke_autocomplete(interaction, name=focused[0].name)

        if isinstance(command, Group) and interaction.data.options is not None:
            child_2 = command.commands.get(interaction.data.options[0].name)
            if (
                isinstance(child_2, Command)
                and interaction.data.options[0].options is not None
            ):
                options = interaction.data.options[0].options
                focused = list(filter(lambda x: x.focused == True, options))
                if not focused:
                    raise ValueError("No focus items! Probably this is a discord bug.")
                return await child_2.invoke_autocomplete(
                    interaction, name=focused[0].name
                )
            if isinstance(child_2, Group) and interaction.data.options[0].options:
                child_3 = child_2.commands.get(
                    interaction.data.options[0].options[0].name
                )
                if (
                    isinstance(child_3, Command)
                    and interaction.data.options[0].options[0].options is not None
                ):
                    options = interaction.data.options[0].options[0].options
                    focused = list(filter(lambda x: x.focused == True, options))
                    if not focused:
                        raise ValueError(
                            "No focus items! Probably this is a discord bug."
                        )
                    return await child_3.invoke_autocomplete(
                        interaction, name=focused[0].name
                    )

    async def _handle_message_component(self, request: Request) -> Any:
        """
        Handles a message component request.

        Parameters
        ----------
        request: :class:`Request`
            The request object.

        Returns
        -------
        The response object.
        """
        payload: dict = await request.json()
        interaction = Interaction(request=request, data=payload)
        if interaction.data and isinstance(interaction.data, MessageComponentData):
            comp = self.client._components.get(interaction.data.custom_id)
            if comp:
                callback = comp._callback
                if callback is None:
                    return
                try:
                    return await callback(interaction)
                except Exception as e:
                    return await comp.view.on_error(interaction, e)

    async def _handle_modal_submit(self, request: Request):
        """
        Handles a modal submit request.

        Parameters
        ----------
        request: :class:`Request`
            The request object.

        Returns
        -------
        The response object.
        """
        payload: dict = await request.json()
        interaction = Interaction(request=request, data=payload)
        if interaction.data is not None and isinstance(
            interaction.data, ModalSubmitData
        ):
            modal = self.client._modals.get(interaction.data.custom_id)
            if modal:
                await modal._invoke(interaction)

    async def handle_interactions(self, request: Request):
        """
        The function is meant to handle interactions posted by
        Discord in the "/interactions" route specified in the bot instance.

        Parameters
        ----------
        request: (Request)
            The request object.

        Returns
        -------
        (Response)
        """
        signature = request.headers["X-Signature-Ed25519"]
        timestamp = request.headers["X-Signature-Timestamp"]
        if (
            signature is None
            or timestamp is None
            or not self.verify_key(await request.body(), signature, timestamp)
        ):
            return Response(content="Bad Signature", status_code=401)

        payload: dict = await request.json()
        interaction = Interaction(request=request, data=payload)
        self.client.dispatch(
            "interaction_create",
            interaction,
            payload=payload,
        )
        if payload["type"] == InteractionType.PING.value:
            return JSONResponse({"type": InteractionResponseType.PONG.value})
        if payload["type"] == InteractionType.APPLICATION_COMMAND.value:
            await self._handle_command(request)
        elif payload["type"] == InteractionType.APPLICATION_COMMAND_AUTOCOMPLETE.value:
            await self._handle_autocomplete(request)
        elif payload["type"] == InteractionType.MESSAGE_COMPONENT.value:
            await self._handle_message_component(request)
        elif payload["type"] == InteractionType.MODAL_SUBMIT.value:
            await self._handle_modal_submit(request)
        return JSONResponse({"ack": InteractionResponseType.PONG.value})



from __future__ import annotations

import uuid

from typing import Any, TYPE_CHECKING
from ..enums import ComponentType, TextInputStyle
from .component import Component
from ..models import ModalSubmitData

if TYPE_CHECKING:
    from ..models import Interaction
    from typing_extensions import Self


__all__ = ("Modal", "TextInput")


class Modal:
    """
    Represents a UI Modal dialog.

    Parameters
    ----------
    title: :class:`str`
        The title of the modal.
        Must be 45 characters or fewer.
    custom_id: :class:`str`
        The custom ID of the modal. If not provided, a random UUID will be generated.
        Must be 100 characters or fewer.

    Attributes
    ----------
    children: :class:`list[TextInput]`
        A list of :class:`TextInput` components.
    """

    def __init__(self, title: str, custom_id: str | None = None) -> None:
        self._title = title
        self._custom_id = custom_id or str(uuid.uuid4())
        self._children: list[TextInput] = list()

        if len(title) > 45:
            raise ValueError("Modal title must be 45 characters or fewer.")

        if len(self._custom_id) > 100:
            raise ValueError("Modal custom_id must be 100 characters or fewer.")

        if len(self.children) > 5:
            raise ValueError("Modal cannot have more than 5 children.")

    def add_item(self, item: TextInput) -> Self:
        self._children.append(item)
        return self

    @property
    def title(self) -> str:
        return self._title

    @property
    def custom_id(self) -> str:
        return self._custom_id

    @property
    def children(self) -> list[TextInput]:
        return self._children

    async def on_error(self, interaction: Interaction, exception: Exception):
        ...

    async def _invoke(self, interaction: Interaction):
        assert isinstance(
            interaction.data, ModalSubmitData
        ), "Invalid interaction recived."
        inputs = [t for r in interaction.data.components for t in r.components]
        for input in inputs:
            item = list(
                filter(lambda x: x.custom_id == input.custom_id, self.children)
            )[0]
            if not item:
                return await self.on_error(
                    interaction,
                    Exception(
                        f"Modal interaction referencing unknown item custom_id {input.custom_id!r}. Discarding"
                    ),
                )
            item.value = input.value
        try:
            await self.on_submit(interaction)
        except Exception as e:
            await self.on_error(interaction, e)

    async def on_submit(self, interaction: Interaction):
        ...

    def __repr__(self) -> str:
        return f"<Modal title={self.title!r}>"

    def to_dict(self) -> dict[str, Any]:
        """
        Converts a :class:`Modal` into a dict.

        Returns
        -------
        dict[str, Any]
        """
        base = {
            "title": self.title,
            "custom_id": self.custom_id,
            "components": [
                {"type": ComponentType.ACTION_ROW.value, "components": [t.to_dict()]}
                for t in self.children[:5]
            ],
        }
        return base


class TextInput(Component):
    """
    Represents a text input component.

    Parameters
    ----------
    label: :class:`str`
        The text displayed to the top of the text field.
        Must be 45 characters or fewer.
    style: :class:`TextInputStyle`
        The style of the text field.
    placeholder: :class:`str`
        The placeholder text displayed when the text field is empty.
        Must be 100 characters or fewer.
    custom_id: :class:`str`
        The custom ID of the text input. If not provided, a random UUID will be generated.
    disabled: :class:`bool`
        Whether the text field is disabled.
    min_length: :class:`int`
        The minimum number of characters that must be entered.
        Defaults to 0 and must be less than 4000.
    max_length: :class:`int`
        The maximum number of characters that can be entered.
        Must be between 1 and 4000.
    required: :class:`bool`
        Whether the text field is required.
    value: :class:`str`
        Pre-fills the input text field with this value.
        Must be 4000 characters or fewer.
    """

    def __init__(
        self,
        label: str,
        style: TextInputStyle = TextInputStyle.short,
        placeholder: str | None = None,
        custom_id: str | None = None,
        disabled: bool | None = None,
        min_length: int | None = None,
        max_length: int | None = None,
        required: bool | None = None,
        value: str | None = None,
    ) -> None:
        super().__init__(ComponentType.TEXT_INPUT, custom_id, disabled)
        self.label = label
        self.style = style
        self.min_length = min_length
        self.max_length = max_length
        self.required = required
        self.value = value
        self.placeholder = placeholder
        if len(self.label) > 45:
            raise ValueError("Label must be 45 characters or fewer.")
        if self.value is not None and len(self.value) > 4000:
            raise ValueError("Value must be 4000 characters or fewer.")

        if self.min_length is not None and self.max_length is not None:
            if self.min_length > self.max_length:
                raise ValueError("Min length must be less than max length.")

        if self.min_length is not None:
            if self.min_length < 1:
                raise ValueError("Min length must be greater than 0.")
        if self.max_length is not None:
            if self.max_length < 1:
                raise ValueError("Max length must be greater than 0.")

    def __repr__(self) -> str:
        return f"<TextInput label={self.label!r}>"

    def to_dict(self) -> dict[str, Any]:
        """
        Converts a :class:`TextInput` to a dict.

        Returns
        -------
        dict[str, Any]
        """
        base = super().to_dict()
        base.update({"label": self.label, "style": self.style.value})
        if self.min_length is not None:
            base["min_length"] = self.min_length
        if self.max_length is not None:
            base["max_length"] = self.max_length
        if self.required is not None:
            base["required"] = self.required
        if self.value is not None:
            base["value"] = self.value
        if self.placeholder is not None:
            base["placeholder"] = self.placeholder
        return base
